from dp.launching.report import Report, ReportSection, ChartReportElement
from typing import List
import glob
import os

from dflow_galaxy.core.util import parse_string_array
import fire


def gen_report(lammps_output_dir: str,
               output_dir: str):
    os.makedirs(output_dir, exist_ok=True)
    sections = []
    fes_file = os.path.join(lammps_output_dir, 'fes.dat')

    element = ChartReportElement(
        title='FES plot',
        options=_gen_fes_echart(fes_file),
    )
    sections.append(ReportSection(
        title='FES',
        ncols=1,
        elements=[element]
    ))

    # write report
    report = Report(title='DynaCat MD', sections=sections)
    report.save(output_dir)


def _gen_fes_echart(f: str):
    header, rows = _load_fes_data(f)

    series = []
    for i, col_name in enumerate(header[1:]):
        series.append({
            'name': col_name,
            'type': 'line',
            'data': [row[i] for row in rows]

        })

    echart = {
        'tooltip': {
            'trigger': 'axis'
        },
        'legend': {
            'data': header[1:],
        },
        'xAxis': {
            'type': 'category',  # FIXME: should use value type
            'name': header[0],
            'data': [row[0] for row in rows],
        },
        'yAxis': {
            'type': 'value',
            'name': 'y-axis'
        },
        'series': series,
    }
    return echart


def _load_fes_data(f: str):
    """
    #! FIELDS d1 file.free der_d1
    #! SET min_d1 0.273758
    #! SET max_d1 6.8191
    #! SET nbins_d1  94
    #! SET periodic_d1 false
    0.273758000    0.998068539   -0.000000000
    0.344138022    0.997859772   -0.008132961
    0.414518043    0.996330768   -0.040475018
    0.484898065    0.991143958   -0.119637947
    0.555278086    0.977356216   -0.288794464

    get fields from first row, ignore other comments, and parse data
    """
    field_prefix = '#! FIELDS'
    header = None
    rows = []
    with open(f, encoding='utf-8') as fp:
        for line in fp:
            line = line.strip()
            if not line:
                continue
            if line.startswith(field_prefix):
                header = parse_string_array(line[len(field_prefix):])
            elif line.startswith('#'):
                continue
            else:
                rows.append(parse_string_array(line, dtype=float))
    assert header is not None
    return header, rows


if __name__ == '__main__':
    fire.Fire({
        'gen_report': gen_report,
    })



from dataclasses import dataclass
from typing import List, Tuple, Optional
from itertools import groupby
from pathlib import Path
import glob

from ai2_kit.domain.deepmd import make_deepmd_task_dirs
from ai2_kit.core.util import cmd_with_checkpoint as cmd_cp, load_text
from ai2_kit.domain.constant import DP_INPUT_FILE, DP_ORIGINAL_MODEL, DP_FROZEN_MODEL

from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core.dispatcher import BaseApp, PythonApp, create_dispatcher, ExecutorConfig
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import bash_iter_ls_slice, safe_ln, bash_ln_cmd, bash_inspect_dir, inspect_dir
from dflow_galaxy.core.log import get_logger
from dflow_galaxy.core import types

from dflow import argo_range
import dpdata

from .lib import LabelApp

logger = get_logger(__name__)

INIT_DATASET_DIR = './init-dataset'
ITER_DATASET_DIR = './iter-dataset'

class DeepmdApp(BaseApp):
    dp_cmd: str = 'dp'
    concurrency: int = 4


class DeepmdConfig(BaseModel):
    model_num: int = 4
    init_dataset: List[str] = []
    input_template: dict = {}
    compress_model: bool = False
    ignore_error: bool = False


@dataclass(frozen=True)
class UpdateDatasetArgs:
    label_dir: types.InputArtifact
    iter_dataset_dir: types.OutputArtifact


class UpdateDatasetFn:
    def __init__(self, config: DeepmdConfig, iter_str: str, label_app: Optional[LabelApp], type_map: List[str]):
        self.config = config
        self.iter_str = iter_str
        self.label_app = label_app
        self.type_map = type_map

    def __call__(self, args: UpdateDatasetArgs):
        dp_sys_list:List[Tuple[str, dpdata.LabeledSystem]] = []

        # parse label data
        if self.label_app == 'cp2k':
            cp2k_dirs = glob.glob(f'{args.label_dir}/tasks/*/persist')
            for cp2k_dir in cp2k_dirs:
                try:
                    ancestor = load_text(f'{cp2k_dir}/ANCESTOR')
                    dp_sys = dpdata.LabeledSystem(f'{cp2k_dir}/output', fmt='cp2k/output', type_map=self.type_map)
                    if dp_sys is not None and len(dp_sys) > 0:
                        dp_sys_list.append((ancestor,dp_sys))
                    else:
                        logger.warn(f'Ignore empty dp system: {cp2k_dir}')
                except Exception as e:
                    logger.exception(f'Failed to load cp2k output: {cp2k_dir}')
                    if not self.config.ignore_error:
                        raise e
        else:
            raise ValueError(f'Unsupported label app: {self.label_app}')

        # dump data
        dataset_dir = Path(args.iter_dataset_dir) / self.iter_str
        dataset_dir.mkdir(parents=True, exist_ok=True)

        # group dataset by ancestor
        dp_sys_list = sorted(dp_sys_list, key=lambda x: x[0])  # sorted by ancestor
        for ancestor, group in groupby(dp_sys_list, key=lambda x: x[0]):
            group = list(group)
            assert ancestor, 'ancestor should not be empty'
            dataset = group[0][1]
            for _a, dp_sys in group[1:]:
                dataset += dp_sys
            dataset.to_deepmd_npy(dataset_dir / ancestor, set_size=len(dataset), type_map=self.type_map)  # type: ignore


@dataclass(frozen=True)
class SetupDeepmdTasksArgs:
    init_dataset_dir: types.InputArtifact
    iter_dataset_dir: types.InputArtifact

    work_dir: types.OutputArtifact


class SetupDeepmdTaskFn:
    def __init__(self, config: DeepmdConfig, type_map: List[str]):
        self.config = config
        self.type_map = type_map

    def __call__(self, args: SetupDeepmdTasksArgs):
        # dflow didn't provide a unified file namespace,
        # so we have to link dataset to a fixed path and use relative path to access it
        safe_ln(args.init_dataset_dir, INIT_DATASET_DIR)
        inspect_dir(INIT_DATASET_DIR)
        safe_ln(args.iter_dataset_dir, ITER_DATASET_DIR)
        inspect_dir(ITER_DATASET_DIR)

        train_dataset_dirs = [ f'{INIT_DATASET_DIR}/{ds}' for ds in self.config.init_dataset]
        train_dataset_dirs.extend(glob.glob(f'{ITER_DATASET_DIR}/*/*'))

        make_deepmd_task_dirs(input_template=self.config.input_template,
                              model_num=self.config.model_num,
                              train_systems=train_dataset_dirs,
                              type_map=self.type_map,
                              base_dir=f'{args.work_dir}/tasks',
                              # TODO: support the following parameters
                              isolate_outliers=False,
                              validation_systems=[],
                              outlier_systems=[],
                              outlier_weight=-1.0,
                              dw_input_template=None,
                              )


@dataclass(frozen=True)
class RunDeepmdTrainingArgs:
    slice_idx: types.InputParam[types.SliceIndex]

    init_dataset_dir: types.InputArtifact
    iter_dataset_dir: types.InputArtifact

    work_dir: types.InputArtifact
    persist_dir: types.OutputArtifact


class RunDeepmdTrainingFn:
    def __init__(self, config: DeepmdConfig, context: DeepmdApp):
        self.config = config
        self.context = context

    def __call__(self, args: RunDeepmdTrainingArgs):
        """generate bash script to run deepmd training commands"""
        c = self.context.concurrency

        script = [
            f'mkdir -p {args.persist_dir}',
            bash_inspect_dir(args.work_dir),
            f"pushd {args.work_dir}",
            bash_iter_ls_slice(
                'tasks/*/', opt='-d', n=c, i=args.slice_idx, it_var='ITEM',
                script=[
                    '# dp train',
                    'pushd $ITEM',
                    'mv persist/* . || true  # recover checkpoint',
                    bash_ln_cmd(args.init_dataset_dir, INIT_DATASET_DIR),
                    bash_ln_cmd(args.iter_dataset_dir, ITER_DATASET_DIR),
                    '',
                    self._build_dp_train_script(),
                    '',
                    '# persist result',
                    f'PERSIST_DIR={args.persist_dir}/$ITEM/persist/',
                    'mkdir -p $PERSIST_DIR',
                    'rm model.ckpt* || true',
                    'mv *.* $PERSIST_DIR',
                    'popd',
                ]
            ),
            'popd',
        ]
        return script

    def _build_dp_train_script(self):
        dp_cmd = self.context.dp_cmd
        train_cmd = f'{dp_cmd} train {DP_INPUT_FILE}'
        # TODO: handle restart, initialize from previous model, support pretrain model
        script = [
            cmd_cp(train_cmd, 'dp-train.done'),
            cmd_cp(f'{dp_cmd} freeze -o {DP_ORIGINAL_MODEL}', 'dp-freeze.done'),
        ]
        # compress (optional) and frozen model
        if self.config.compress_model:
            freeze_cmd = f'{dp_cmd} compress -i {DP_ORIGINAL_MODEL} -o {DP_FROZEN_MODEL}'
        else:
            freeze_cmd = f'mv {DP_ORIGINAL_MODEL} {DP_FROZEN_MODEL}'
        script.append(cmd_cp(freeze_cmd, 'dp-compress.done'))

        return '\n'.join(script)


def provision_deepmd(builder: DFlowBuilder, ns: str, /,
                     config: DeepmdConfig,
                     executor: ExecutorConfig,
                     deepmd_app: DeepmdApp,
                     python_app: PythonApp,
                     work_dir_url: str,

                     label_app: Optional[LabelApp],
                     label_dir_url: Optional[str],

                     init_dataset_url: str,
                     iter_dataset_url: str,
                     iter_str: str,
                     type_map: List[str],
                     ):
    if label_app and label_dir_url:
        update_dataset_fn = UpdateDatasetFn(config, iter_str=iter_str,
                                            label_app=label_app, type_map=type_map)
        update_dataset_step = builder.make_python_step(update_dataset_fn, uid=f'{ns}-update-dataset',
                                                       setup_script=python_app.setup_script,
                                                       executor=create_dispatcher(executor, python_app.resource))(
            UpdateDatasetArgs(
                label_dir=label_dir_url,
                iter_dataset_dir=iter_dataset_url,
            )

        )
        builder.add_step(update_dataset_step)

    setup_tasks_fn = SetupDeepmdTaskFn(config, type_map)
    setup_tasks_step = builder.make_python_step(setup_tasks_fn, uid=f'{ns}-setup-task',
                                                setup_script=python_app.setup_script,
                                                executor=create_dispatcher(executor, python_app.resource))(
        SetupDeepmdTasksArgs(
            init_dataset_dir=init_dataset_url,
            iter_dataset_dir=iter_dataset_url,
            work_dir=work_dir_url,
        )
    )
    run_training_fn = RunDeepmdTrainingFn(config=config, context=deepmd_app)
    run_training_step = builder.make_bash_step(run_training_fn, uid=f'{ns}-run-training',
                                               setup_script=deepmd_app.setup_script,
                                               with_param=argo_range(deepmd_app.concurrency),
                                               executor=create_dispatcher(executor, deepmd_app.resource))(
        RunDeepmdTrainingArgs(
            slice_idx="{{item}}",
            init_dataset_dir=init_dataset_url,
            iter_dataset_dir=iter_dataset_url,
            work_dir=work_dir_url,
            persist_dir=work_dir_url,
        )
    )

    builder.add_step(setup_tasks_step)
    builder.add_step(run_training_step)



from typing import List, Mapping, Optional, Any
from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core import dispatcher
from ai2_kit.core.artifact import Artifact

from .domain import deepmd, lammps, model_devi, cp2k


class GeneralConfig(BaseModel):
    type_map: List[str]
    mass_map: List[float]


class AppsConfig(BaseModel):
    python: dispatcher.PythonApp
    deepmd: Optional['deepmd.DeepmdApp'] = None
    lammps: Optional['lammps.LammpsApp'] = None
    cp2k: Optional['cp2k.Cp2kApp'] = None


class Orchestration(BaseModel):
    deepmd: Optional[str] = None
    lammps: Optional[str] = None
    model_devi: Optional[str] = None
    cp2k: Optional[str] = None


class TeslaExecutorConfig(dispatcher.ExecutorConfig):
    apps: AppsConfig


class LabelConfig(BaseModel):
    cp2k: Optional['cp2k.Cp2kConfig']


class TrainConfig(BaseModel):
    deepmd: Optional['deepmd.DeepmdConfig']


class ExploreConfig(BaseModel):
    lammps: Optional['lammps.LammpsConfig']


class ScreenConfig(BaseModel):
    model_devi: Optional['model_devi.ModelDeviConfig']


class UpdateConfig(BaseModel):
    until_iter: int
    patch: dict


class WorkflowConfig(BaseModel):
    general: GeneralConfig
    label: LabelConfig
    train: TrainConfig
    explore: ExploreConfig
    screen: ScreenConfig
    update: Optional[UpdateConfig] = None


class TeslaConfig(BaseModel):
    executors: Mapping[str, TeslaExecutorConfig]
    orchestration: Orchestration
    datasets: Mapping[str, Artifact]
    workflow: Any

    def init(self):
        for k, v in self.datasets.items():
            assert 'ancestor' not in v.attrs, f'ancestor already exists: {v.attrs}'
            v.attrs['ancestor'] = k



from dataclasses import dataclass
from dflow_galaxy.core import types, dispatcher
from dflow_galaxy.core.dflow_builder import DFlowBuilder
import shutil
import os


@dataclass(frozen=True)
class RunCp2kArgs:
    input_dir : types.InputArtifact
    output_dir: types.OutputArtifact


class RunCp2kFn:
    def __init__(self, cp2k_script: str):
        self.cp2k_script = cp2k_script

    def __call__(self, args: RunCp2kArgs):
        """
        bash step to run cp2k aimd task
        """
        script = [
            # guess cp2k data dir
            '[[ -z "${CP2K_DATA_DIR}" ]] && export CP2K_DATA_DIR="$(dirname "$(which cp2k || which cp2k.psmp)")/../../data" || true',
            f'cd {args.input_dir}',
            self.cp2k_script,
            f'mkdir -p {args.output_dir}',
            f'mv * {args.output_dir}',
        ]
        return script


def run_cp2k_workflow(input_dir: str,
                      out_dir: str,
                      cp2k_image: str,
                      cp2k_device_model: str,
                      cp2k_script: str):

    # bohrium dispatcher will be configured in bohrium.config
    # so here we just leave it empty
    bohrium_config = dispatcher.BohriumConfig()

    # start to build workflow
    dflow_builder = DFlowBuilder('cp2k', s3_prefix='cp2k')

    # setup and add cp2k step to workflow
    cp2k_res = dispatcher.Resource(
        bohrium=dispatcher.BohriumInputData(
            image_name=cp2k_image,
            scass_type=cp2k_device_model,
            disk_size=100,
            job_name='cp2k_lightning',
        )
    )

    dflow_builder.s3_upload(input_dir, 'cp2k_input')
    cp2k_executor = dispatcher.create_bohrium_dispatcher(bohrium_config, cp2k_res)
    cp2k_fn = RunCp2kFn(cp2k_script=cp2k_script)
    cp2k_step = dflow_builder.make_bash_step(cp2k_fn, executor=cp2k_executor)(
        RunCp2kArgs(input_dir='s3://./cp2k_input', output_dir='s3://./cp2k_output')
    )
    dflow_builder.add_step(cp2k_step)

    # run workflow
    dflow_builder.run()

    # download artifacts to out_dir
    dflow_builder.s3_download('cp2k_output')
    shutil.unpack_archive('cp2k_output', out_dir, format='gztar')
    return os.path.join(out_dir, 'output_dir')



from typing import List, Optional, Mapping, Any, Literal
from dataclasses import dataclass
from copy import deepcopy
import glob
import os

from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core.dispatcher import BaseApp, PythonApp, create_dispatcher, ExecutorConfig
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import bash_iter_ls_slice, safe_ln, bash_ln_cmd, inspect_dir, bash_inspect_dir
from dflow_galaxy.core import types

from ai2_kit.domain.lammps import make_lammps_task_dirs, FepOptions
from ai2_kit.domain.constant import DP_FROZEN_MODEL
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.util import cmd_with_checkpoint as cmd_cp, dump_text

from dflow import argo_range


from .lib import resolve_artifact


MODEL_DIR = './mlp-models'
SYSTEM_DIR = './systems'


class LammpsApp(BaseApp):
    lammps_cmd: str = 'lmp'
    concurrency: int = 5


class LammpsConfig(BaseModel):
    systems: List[str]

    nsteps: int
    no_pbc: bool = False
    timestep: float = 0.0005
    sample_freq: int = 100

    product_vars: Mapping[str, List[Any]]
    broadcast_vars: Mapping[str, Any] = {}

    template_vars: Mapping[str, Any] = {}
    """
    input_template may provide extra injection points for user to inject custom settings.
    Those value could be set here.

    Those vars can be referenced in the LAMMPS input template as $$VAR_NAME.
    """

    input_template: Optional[str] = None
    plumed_config: Optional[str] = None
    ensemble: Optional[Literal['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve', 'csvr']] = None
    ignore_error: bool = False


@dataclass(frozen=True)
class SetupLammpsTasksArgs:
    model_dir: types.InputArtifact
    system_dir: types.InputArtifact
    work_dir: types.OutputArtifact


class SetupLammpsTasksFn:
    def __init__(self, config: LammpsConfig,
                 type_map: List[str],
                 mass_map: List[float],
                 systems: Mapping[str, Artifact]):
        self.config = config
        self.type_map = type_map
        self.mass_map = mass_map
        self.systems = systems

    def __call__(self, args: SetupLammpsTasksArgs):
        # dflow didn't provide a unified file namespace,
        # so we have to link dataset to a fixed path and use relative path to access it
        safe_ln(args.model_dir, MODEL_DIR)
        inspect_dir(MODEL_DIR)
        safe_ln(args.system_dir, SYSTEM_DIR)
        inspect_dir(SYSTEM_DIR)

        # resolve input data
        data_files: List[ArtifactDict] = []
        for k in self.config.systems:
            v = deepcopy(self.systems[k])  # avoid side effect
            v.url = os.path.join(SYSTEM_DIR, k)
            data_files.extend(resolve_artifact(v))

        # resolve model files
        search_pattern = f'{MODEL_DIR}/tasks/*/persist/{DP_FROZEN_MODEL}'
        model_files = glob.glob(search_pattern)
        assert model_files, f'no model files found in {search_pattern}'

        # handle default value
        default_vars = {
            'TAU_T': [0.1],
            'TAU_P': [0.5],
            'TIME_CONST': [0.1],
        }
        for k, v in default_vars.items():
            if k not in self.config.product_vars and k not in self.config.broadcast_vars:
                self.config.broadcast_vars[k] = v  # type: ignore

        _base_dir, task_dirs = make_lammps_task_dirs(
            combination_vars=self.config.product_vars,
            broadcast_vars=self.config.broadcast_vars,
            data_files=data_files,
            dp_models={'': model_files},
            n_steps=self.config.nsteps,
            timestep=self.config.timestep,
            sample_freq=self.config.sample_freq,
            no_pbc=self.config.no_pbc,
            ensemble=self.config.ensemble,
            input_template=self.config.input_template,
            plumed_config=self.config.plumed_config,
            extra_template_vars=self.config.template_vars,
            type_map=self.type_map,
            mass_map=self.mass_map,
            work_dir=args.work_dir,

            # TODO: support more feature in the future
            mode='default',
            type_alias={},
            dp_modifier=None,
            dp_sel_type=None,
            preset_template='default',
            n_wise=0,
            fix_statement=None,
            rel_path=True,
            fep_opts=FepOptions(),
        )
        # write ancestor to the task dirs
        for task_dir in task_dirs:
            path = os.path.join(task_dir['url'], 'ANCESTOR')
            dump_text(task_dir['attrs']['ancestor'], path)


@dataclass(frozen=True)
class RunLammpsTasksArgs:
    slice_idx: types.InputParam[types.SliceIndex]
    model_dir: types.InputArtifact
    work_dir: types.InputArtifact
    persist_dir: types.OutputArtifact


class RunLammpsTasksFn:
    def __init__(self, config: LammpsConfig, context: LammpsApp):
        self.config = config
        self.context = context

    def __call__(self, args: RunLammpsTasksArgs):
        c = self.context.concurrency

        script = [
            f'mkdir -p {args.persist_dir} && touch {args.persist_dir}/.placeholder',
            bash_inspect_dir(args.work_dir),
            f"pushd {args.work_dir}",
            bash_iter_ls_slice(
                'tasks/*/', opt='-d', n=c, i=args.slice_idx, it_var='ITEM',
                script=[
                    '# run lammps',
                    'pushd $ITEM',
                    bash_ln_cmd(args.model_dir, MODEL_DIR),
                    'mv persist/* . || true  # restore previous state',
                    '',
                    self._build_lammps_cmd(),
                    '',
                    '# persist result',
                    f'PERSIST_DIR={args.persist_dir}/$ITEM/persist/',
                    'mkdir -p $PERSIST_DIR',
                    'mv *.done traj model_devi.out ANCESTOR $PERSIST_DIR',
                    'popd',
                ]
            ),
            'popd',
        ]
        return script

    def _build_lammps_cmd(self):
        lmp_cmd = f'{self.context.lammps_cmd} -i lammps.input'
        cmd = f'''if [ -f md.restart.* ]; then {lmp_cmd} -v restart 1; else {lmp_cmd} -v restart 0; fi'''
        return cmd_cp(cmd, 'lammps.done', ignore_error=self.config.ignore_error)


def provision_lammps(builder: DFlowBuilder, ns: str, /,
                     config: LammpsConfig,
                     executor: ExecutorConfig,
                     lammps_app: LammpsApp,
                     python_app: PythonApp,

                     systems_url: str,
                     mlp_model_url: str,
                     work_dir_url: str,
                     type_map: List[str],
                     mass_map: List[float],
                     systems: Mapping[str, Artifact],
                     ):
    setup_tasks_fn = SetupLammpsTasksFn(config, type_map=type_map, mass_map=mass_map, systems=systems)
    setup_tasks_step = builder.make_python_step(setup_tasks_fn, uid=f'{ns}-setup-task',
                                                setup_script=python_app.setup_script,
                                                executor=create_dispatcher(executor, python_app.resource))(
        SetupLammpsTasksArgs(
            model_dir=mlp_model_url,
            system_dir=systems_url,
            work_dir=work_dir_url,
        )
    )

    run_tasks_fn = RunLammpsTasksFn(config, lammps_app)
    run_tasks_step = builder.make_bash_step(run_tasks_fn, uid=f'{ns}-run-task',
                                            setup_script=lammps_app.setup_script,
                                            with_param=argo_range(lammps_app.concurrency),
                                            executor=create_dispatcher(executor, lammps_app.resource))(
        RunLammpsTasksArgs(
            slice_idx='{{item}}',
            model_dir=mlp_model_url,
            work_dir=work_dir_url,
            persist_dir=work_dir_url,
        )
    )

    builder.add_step(setup_tasks_step)
    builder.add_step(run_tasks_step)



from .pydantic import BaseModel
from pydantic import Field


from typing import Optional
from urllib.parse import urlparse
import os


class BohriumInputData(BaseModel):
    image_name: str
    scass_type: str = 'c2_m4_cpu'
    job_type: str = 'container'
    platform: str = 'ali'
    job_name: str = 'by_dflow_galaxy'
    disk_size: int = 20  # in GB


class Resource(BaseModel):
    queue: Optional[str] = None
    image: Optional[str] = None
    sub_path: str = '.'

    nodes: int = 1
    cpu_per_node: int = 1
    gpu_per_node: int = 0

    # use this for job running on bohrium
    bohrium: Optional[BohriumInputData] = None


    def get_resource_dict(self):
        return {
            'number_node': self.nodes,
            'cpu_per_node': self.cpu_per_node,
            'gpu_per_node': self.gpu_per_node,
        }


class BohriumConfig(BaseModel):
    email: Optional[str] = None
    password: Optional[str] = None
    project_id: Optional[str] = None


class HpcConfig(BaseModel):
    class SlurmConfig(BaseModel):
        ...
    class LsfConfig(BaseModel):
        ...
    class PBSConfig(BaseModel):
        ...

    url: str
    """
    SSH URL to connect to the HPC, for example: `john@hpc-login01`
    """
    key_file: Optional[str] = None
    """
    Path to the private key file for SSH connection
    """
    slurm: Optional[SlurmConfig] = None
    lsf: Optional[LsfConfig] = None
    pbs: Optional[PBSConfig] = None
    base_dir: str = '.'
    clean: bool = False

    def get_context_type(self):
        if self.slurm:
            return 'Slurm'
        if self.lsf:
            return 'LSF'
        if self.pbs:
            return 'PBS'
        raise ValueError('At least one of slurm, lsf or pbs should be provided')


class ExecutorConfig(BaseModel):
    hpc: Optional[HpcConfig] = None
    bohrium: Optional[BohriumConfig] = None


def create_dispatcher(config: ExecutorConfig, resource: Resource):
    """
    Create a dispatcher executor based on the configuration
    """
    if config.hpc:
        return create_hpc_dispatcher(config.hpc, resource)
    elif config.bohrium:
        return create_bohrium_dispatcher(config.bohrium, resource)
    raise ValueError('At least one of hpc or bohrium should be provided')


def create_bohrium_dispatcher(config: BohriumConfig, resource: Resource):
    if resource.bohrium is None:
        raise ValueError('Bohrium resource is required when using Bohrium dispatcher')
    from dflow.plugins.dispatcher import DispatcherExecutor
    password = os.environ.get('BOHRIUM_PASSWORD') or config.password

    # remote profile can also be config via global bohrium.config
    remote_profile = {}
    if config.email:
        remote_profile['email'] = config.email

    if config.project_id:
        # I don't know which one is correct, so I set both
        remote_profile['program_id'] = config.project_id
        remote_profile['project_id'] = config.project_id
    if password:
        remote_profile['password'] = password

    remote_profile['input_data'] = resource.bohrium.dict()
    machine_dict = {
        'batch_type': 'Bohrium',
        'context_type': 'Bohrium',
        'remote_profile': remote_profile,
    }
    return DispatcherExecutor(
        machine_dict=machine_dict,
        # the following are not used in Bohrium dispatcher
        image=resource.image,
        resources_dict= resource.get_resource_dict(),
    )


def create_hpc_dispatcher(config: HpcConfig, resource: Resource):
    from dflow.plugins.dispatcher import DispatcherExecutor
    url = urlparse(config.url)
    assert url.scheme == 'ssh', 'Only SSH is supported for HPC dispatcher'
    assert url.username, 'Username is required in the URL'
    assert os.path.isabs(config.base_dir), 'Base directory must be an absolute path'
    remote_root = os.path.normpath(
        os.path.join(config.base_dir, resource.sub_path))

    remote_profile = { }
    if config.key_file:
        remote_profile['key_filename'] = config.key_file
    machine_dict = {
        "batch_type": config.get_context_type(),
        "context_type": "SSHContext",
        'remote_profile': remote_profile,
    }

    return DispatcherExecutor(
        host=url.hostname or 'localhost',
        private_key_file=config.key_file,  # type: ignore
        username=url.username,
        port=url.port or 22,
        clean=config.clean,
        machine_dict=machine_dict,
        resources_dict=resource.get_resource_dict(),
        queue_name=resource.queue,
        remote_root=remote_root,
    )


class BaseApp(BaseModel):
    resource: Resource
    setup_script: str = ''


class PythonApp(BaseApp):
    python_cmd: str = 'python3'
    max_worker: int = 4



from typing import List, Optional, Mapping, Any, Literal
from dataclasses import dataclass
from copy import deepcopy
from pathlib import Path
import glob
import os

from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core.dispatcher import BaseApp, PythonApp, create_dispatcher, ExecutorConfig
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import bash_iter_ls_slice, safe_ln, bash_ln_cmd, bash_inspect_dir, inspect_dir
from dflow_galaxy.core import types

from ai2_kit.domain.cp2k import make_cp2k_task_dirs
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.util import cmd_with_checkpoint as cmd_cp, list_sample, load_text, dump_text, ensure_dir

from dflow import argo_range


@dataclass(frozen=True)
class RunReweightingTaskArgs:
    baseline_system_dir: types.InputArtifact
    baseline_colvar_file: types.InputArtifact
    baseline_model_file : Optional[types.InputArtifact]
    target_model_file: types.InputArtifact

    output_dir: types.OutputArtifact


class RunReweightingTask:

    def __init__(self):
        ...

    def __call__(self, args: RunReweightingTaskArgs):
        ...



from typing import Optional
from copy import deepcopy

from ai2_kit.core.util import load_yaml_files, merge_dict
from ai2_kit.core.cmd import CmdGroup

from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import not_none
from dflow_galaxy.core.log import get_logger


logger = get_logger(__name__)


class Config(BaseModel):
    baseline_system_dir: str
    """
    The directory contains the baseline system files in dpdata.System or dpdata.LabeledSystem format.
    """

    baseline_colvar_file: str
    """
    Path to the baseline plumed COLVAR file.
    """

    baseline_model_file: Optional[str] = None
    """
    Path to the baseline DeepMD model file.

    If not provided, the baseline system data should be in dpdata.LabeledSystem format.
    """

    test_model_file: str
    """
    Path to the DeepMD model file for testing.
    """


def build_reweighing_workflow(config: Config,
                              s3_prefix: str,
                              debug: bool = False,
                              name: str = 'reweighting'):

    builder = DFlowBuilder(name=name, s3_prefix=s3_prefix, debug=debug,
                           default_archive=None)



from typing import List, Optional, Mapping, Any, Literal
from dataclasses import dataclass
from copy import deepcopy
from pathlib import Path
import glob
import os

from dflow_galaxy.core.pydantic import BaseModel
from dflow_galaxy.core.dispatcher import BaseApp, PythonApp, create_dispatcher, ExecutorConfig
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import bash_iter_ls_slice, safe_ln, bash_ln_cmd, bash_inspect_dir, inspect_dir
from dflow_galaxy.core import types

from ai2_kit.domain.cp2k import make_cp2k_task_dirs
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.util import cmd_with_checkpoint as cmd_cp, list_sample, load_text, dump_text, ensure_dir

from dflow import argo_range

from .lib import resolve_artifact

SYSTEM_DIR = './system_dir'


class Cp2kApp(BaseApp):
    cp2k_cmd: str = 'cp2k.popt'
    concurrency: int = 5


class Cp2kConfig(BaseModel):
    init_systems: List[str] = []
    input_template: Optional[str] = None
    template_vars: Mapping[str, Any] = {}

    limit: int = 50
    limit_method: Literal['even', 'random', 'truncate'] = 'even'


@dataclass(frozen=True)
class SetupCp2kTasksArgs:
    system_dir: types.InputArtifact
    work_dir: types.OutputArtifact


class SetupCp2kTaskFn:

    def __init__(self, config: Cp2kConfig, systems: Mapping[str, Artifact], init: bool):
        self.config = config
        self.init = init
        self.systems = systems

    def __call__(self, args: SetupCp2kTasksArgs):
        safe_ln(args.system_dir, SYSTEM_DIR)
        inspect_dir(SYSTEM_DIR)

        limit = self.config.limit

        system_files: List[ArtifactDict] = []
        if self.init:
            # handle init systems
            limit = 0  # no limit for init systems
            assert self.config.init_systems, 'init_systems should not be empty for first iteration'
            for k in self.config.init_systems:
                v = deepcopy(self.systems[k])  # avoid side effect
                v.url = os.path.join(SYSTEM_DIR, k)
                system_files.extend(resolve_artifact(v))
        else:
            # handle iter systems
            system_dirs = glob.glob(f'{SYSTEM_DIR}/system/*')  # search pattern is defined by model_devi
            for sys_dir in system_dirs:
                sys_dir = Path(sys_dir)
                ancestor = load_text(sys_dir / 'ANCESTOR')
                a_dict = {
                    'url': str(sys_dir / 'decent.xyz'),
                    'format': 'extxyz',
                    'attrs': deepcopy(self.systems[ancestor].attrs),
                }
                system_files.append(a_dict)  # type: ignore

        assert system_files, 'no system files found'

        task_dir = os.path.join(args.work_dir, 'tasks')
        ensure_dir(task_dir)
        task_dirs = make_cp2k_task_dirs(
            system_files=system_files,
            input_template=self.config.input_template,
            template_vars=self.config.template_vars,
            base_dir=task_dir,
            limit=limit,
            limit_method=self.config.limit_method,
            # not supported yet
            mode='default',
            wfn_warmup_template=None,
            # TODO: type_map is no longer needed, should be fixed in ai2-kit
            type_map=[],
        )
        for task_dir in task_dirs:
            path = os.path.join(task_dir['url'], 'ANCESTOR')
            dump_text(task_dir['attrs']['ancestor'], path)


@dataclass(frozen=True)
class RunCp2kTasksArgs:
    slice_idx: types.InputParam[types.SliceIndex]
    work_dir: types.InputArtifact
    persist_dir: types.OutputArtifact


class RunCp2kTasksFn:
    def __init__(self, config: Cp2kConfig, context: Cp2kApp):
        self.config = config
        self.context = context

    def __call__(self, args: RunCp2kTasksArgs):
        c = self.context.concurrency

        script = [
            f'mkdir -p {args.persist_dir} && touch {args.persist_dir}/.placeholder',
            f"pushd {args.work_dir}",
            bash_inspect_dir('.'),
            bash_iter_ls_slice(
                'tasks/*/', opt='-d', n=c, i=args.slice_idx, it_var='ITEM',
                script=[
                    'pushd $ITEM',
                    'mv persist/* . || true  # recover checkpoint',
                    '',
                    self._build_cp2k_script(),
                    '',
                    '# persist result',
                    f'PERSIST_DIR={args.persist_dir}/$ITEM/persist/',
                    'mkdir -p $PERSIST_DIR',
                    'mv *.done output ANCESTOR $PERSIST_DIR',
                    'popd',
                ]
            ),
            'popd',
        ]
        return script

    def _build_cp2k_script(self):
        cmd = f'''{self.context.cp2k_cmd} -i input.inp &> output'''
        return cmd_cp(cmd, 'cp2k.done')


def provision_cp2k(builder: DFlowBuilder, ns: str, /,
                   config: Cp2kConfig,
                   executor: ExecutorConfig,
                   cp2k_app: Cp2kApp,
                   python_app: PythonApp,

                   system_url: str,
                   work_dir_url: str,

                   init: bool,
                   systems: Mapping[str, Artifact],
                   ):
    setup_tasks_fn = SetupCp2kTaskFn(config, systems=systems, init=init)
    setup_tasks_step = builder.make_python_step(setup_tasks_fn, uid=f'{ns}-setup-task',
                                                setup_script=python_app.setup_script,
                                                executor=create_dispatcher(executor, python_app.resource))(
        SetupCp2kTasksArgs(
            system_dir=system_url,
            work_dir=work_dir_url,
        )
    )

    run_tasks_fn = RunCp2kTasksFn(config, cp2k_app)
    run_tasks_step = builder.make_bash_step(run_tasks_fn, uid=f'{ns}-run-task',
                                            setup_script=cp2k_app.setup_script,
                                            with_param=argo_range(cp2k_app.concurrency),
                                            executor=create_dispatcher(executor, cp2k_app.resource))(
        RunCp2kTasksArgs(
            slice_idx='{{item}}',
            work_dir=work_dir_url,
            persist_dir=work_dir_url,
        )
    )

    builder.add_step(setup_tasks_step)
    builder.add_step(run_tasks_step)



from fire import Fire


class Workflows:

    def tesla(self):
        from dflow_galaxy.workflow.tesla.main import cmd_entry
        return cmd_entry


class CliRoot:
    """
    DFlow Galaxy CLI Tool
    """

    def workflow(self):
        return Workflows()


def main():
    Fire(CliRoot())


if __name__ == '__main__':
    main()


from dp.launching.typing import BaseModel, Field, OutputDirectory, InputFilePath
from dp.launching.typing import Int, String, Enum, Float, Boolean, List, Optional, Dict
from dp.launching.cli import to_runner, default_minimal_exception_handler
from dp.launching.report import Report, ReportSection, ChartReportElement

from dflow_galaxy.app.common import DFlowOptionsMixin, setup_dflow_context, EnsembleOptions
from dflow_galaxy.res import get_res_path
from dflow_galaxy.core.log import get_logger
from dflow_galaxy.core.util import parse_string_array, str_or_none

from dflow_galaxy.workflow.tesla.main import build_tesla_workflow

from ai2_kit.core.util import dump_json, load_text, load_json
from ai2_kit.feat import catalysis as ai2cat

import ase.io
import fire

from pathlib import Path
from uuid import uuid4
import shutil
import glob
import sys
import os

from .report import gen_report

logger = get_logger(__name__)


BH_DEEPMD_DEFAULT = {
    'image_name': 'registry.dp.tech/dptech/dpmd:2.2.8-cuda12.0',
    'scass_type': 'c8_m32_1 * NVIDIA V100',
}
BH_LAMMPS_DEFAULT = BH_DEEPMD_DEFAULT

BH_CP2K_DEFAULT = {
    'image_name': 'registry.dp.tech/dptech/cp2k:11',
    'scass_type': 'c32_m64_cpu',
}

BH_PYTHON_DEFAULT = {
    'image_name': 'registry.dp.tech/dptech/prod-13325/dflow-galaxy:0.1.7-main-07745fb',
    'scass_type': 'c2_m4_cpu',
}


class KvItem(BaseModel):
    key: String = Field(
        description="Key of the item")
    value: String = Field(
        format='multi-line',
        description="Value of the item")


class ExploreItem(BaseModel):
    key: String = Field(
        description="Key of the item")
    value: String = Field(
        description="Value of the item, multiple value should be separated by comma")
    broadcast: Boolean = Field(
        default=False,
        description="Use broadcast instead of full combination")


class DeepmdSettings(BaseModel):

    compress_model: Boolean = Field(
        default=True,
        description="Compress the model after training")

    concurrency: Int = Field(
        default=4,
        description="Number of concurrent run")

    steps: Optional[Int] = Field(
        default=400000,
        description="Number of steps for DeepMD training, leave it empty will use the numb_steps defined in the input template")

    cmd: String = Field(
        default='dp',
        description="Command to run DeepMD, note that it depends on the docker image you used")

    setup_script: String = Field(
        default='',
        format='multi-line',
        description="Setup script for DeepMD simulation, note that it depends on the docker image you used")


class LammpsSetting(BaseModel):
    ensemble: EnsembleOptions = Field(
        default=EnsembleOptions.csvr,
        description='Ensemble of LAMMPS simulation')

    plumed_config: Optional[String] = Field(
        format='multi-line',
        description='Plumed configuration file for metadynamics simulation')

    explore_vars: List[ExploreItem] = Field(
        default=[
           ExploreItem(key='TEMP', value='50,100,200,300,400,600,800,1000', broadcast=False),
           ExploreItem(key='PRES', value='1', broadcast=False),
        ],
        description="Variables for LAMMPS exploration, TEMP, PRES are required, you may also add TAU_T, TAU_P, etc")

    template_vars: List[KvItem] = Field(
        default=[
            KvItem(key='POST_INIT', value='\n'.join([
                'neighbor 1.0 bin',
                'box      tilt large',
            ])),
            KvItem(key='POST_READ_DATA', value='\n'.join([
                'change_box all triclinic'
            ])),
        ],
        description="Template variables for LAMMPS exploration, you may need to modify this according to your system")

    nsteps: Int = Field(
        default=1000,
        description='Number of steps of LAMMPS simulation for the first iteration')

    timestep: Float = Field(
        default=0.0005,
        description='Time step size of LAMMPS simulation in ps')

    sample_freq: Int = Field(
        default=10,
        description='Sampling frequency of LAMMPS simulation')

    no_pbc: Boolean = Field(
        default=False,
        description='Whether to use periodic boundary condition')

    concurrency: Int = Field(
        default=5,
        description="Number of concurrent run")

    cmd: String = Field(
        default='lmp',
        description="Command to run LAMMPS, note that it depends on the docker image you used")

    setup_script: String = Field(
        default='',
        format='multi-line',
        description="Setup script for LAMMPS simulation, note that it depends on the docker image you used")


class ModelDeviation(BaseModel):
    lo: Float = Field(
        default=0.2,
        description="Lower bound of the deviation")

    hi: Float = Field(
        default=0.6,
        description="Upper bound of the deviation")

    metric: String = Field(
        default='max_devi_f',
        description="Metric to measure the deviation")


class Cp2kSettings(BaseModel):

    limit: Int = Field(
        default=50,
        description="Limit of the number of structures to be labeled for each iteration")

    concurrency: Int = Field(
        default=5,
        description="Number of concurrent run")

    cmd: String = Field(
        default='mpirun -np 32 cp2k.popt',
        description="Script to run CP2K simulation, note that it depends on the docker image")

    setup_script: String = Field(
        default = '\n'.join([
            '# guess cp2k data dir',
            '[[ -z "${CP2K_DATA_DIR}" ]] && export CP2K_DATA_DIR="$(dirname "$(which cp2k || which cp2k.psmp)")/../../data" || true',
            'source /opt/cp2k-toolchain/install/setup',
        ]),
        format='multi-line',
        description="Setup script for CP2K simulation, note that it depends on the docker image you used")


class DynacatTeslaArgs(BaseModel, DFlowOptionsMixin):
    deepmd_dataset : InputFilePath = Field(
        title='DeepMD Dataset',
        description="DeepMD in zip or tgz format, for example: deepmd-dataset.tgz")

    deepmd_input_template: Optional[InputFilePath] = Field(
        title='DeepMD Input Template',
        description="Input template file for DeepMD training, in json format, for example: deepmd.json, if not provided, the build-in template will be used")

    lammps_system_file: InputFilePath = Field(
        description="Structure file in xyz format use for LAMMPS simulation, for example h2o.xyz, the xyz file may contain multiple structures")

    cp2k_input_template: InputFilePath = Field(
        description="Input template file for CP2K simulation, for example: cp2k.inp")

    dry_run: Boolean = Field(
        default = True,
        description="Generate configuration file without running the workflow")

    max_iters: Int = Field(
        default = 7,
        description="Maximum iterations of the workflow")

    s3_prefix: Optional[String] = Field(
        description="Specify the S3 prefix of DFlow. By default a random prefix will be used for different jobs. Jobs use the same prefix will share the same working directory, which allow you to inherit the state of previous run.")

    deepmd: DeepmdSettings = Field(
        description="DeepMD settings for training")

    lammps: LammpsSetting = Field(
        description="LAMMPS settings for structure exploration")

    model_deviation: ModelDeviation = Field(
        description="Model deviation settings for screening structures")

    cp2k: Cp2kSettings = Field(
        description="CP2K settings for DFT labeling")

    output_dir : OutputDirectory = Field(
        default='./output',
        description="Output directory of LAMMPS simulation")

    deepmd_resource: Dict[String, String] = Field(
        default=BH_DEEPMD_DEFAULT,
        description="Resource configuration for DeepMD")

    lammps_resource: Dict[String, String] = Field(
        default=BH_LAMMPS_DEFAULT,
        description="Resource configuration for LAMMPS")

    cp2k_resource: Dict[String, String] = Field(
        default=BH_CP2K_DEFAULT,
        description="Resource configuration for CP2K")

    python_resource: Dict[String, String] = Field(
        default=BH_PYTHON_DEFAULT,
        description="Resource configuration for Python")


def launch_app(args: DynacatTeslaArgs) -> int:
    s3_prefix = args.s3_prefix or f'dynacat-{uuid4()}'
    logger.info(f'using s3 prefix: {s3_prefix}')
    tesla_template = get_res_path() / 'dynacat' / 'tesla_template.yml'
    # handle deepmd dataset
    dp_dataset_file = args.deepmd_dataset.get_full_path()
    dp_dataset = _unpack_dpdata(dp_dataset_file, 'init-dataset')
    dp_dataset_config = {}
    for i, d in enumerate(dp_dataset):
        dp_dataset_config[f'dpdata-{i}'] = {
            'url': d,
        }
    logger.info(f'Unpacked dpdata to {dp_dataset}')

    # build config
    shutil.copy(tesla_template, 'tesla-preset.yml')
    executor_config = _get_executor_config(args)
    workflow_config = _get_workflow_config(args, dp_dataset_config)

    # TODO: use yaml to dump config pretty
    dump_json(executor_config, 'tesla-executor.yml')
    dump_json(workflow_config, 'tesla-workflow.yml')

    # copy generated configuration to output
    os.makedirs(args.output_dir, exist_ok=True)
    os.system(f'cp tesla-*.yml {args.output_dir}')

    if args.dry_run:
        logger.info('Skip running workflow due to dry_run is set to True')
        return 0

    setup_dflow_context(args)
    workflow = build_tesla_workflow('tesla-preset.yml', 'tesla-executor.yml', 'tesla-workflow.yml',
                                    s3_prefix=s3_prefix, max_iters=args.max_iters)
    try:
        workflow.run()
    except:
        logger.exception('Failed to run workflow')
        return 1
    finally:
        # reclaim useful data
        dp_dataset_dir = os.path.join(args.output_dir, 'dp-dataset')
        dp_models_dir = os.path.join(args.output_dir, 'dp-models')
        model_devi_dir = os.path.join(args.output_dir, 'model-devi')
        os.makedirs(dp_dataset_dir, exist_ok=True)
        os.makedirs(dp_models_dir, exist_ok=True)
        os.makedirs(model_devi_dir, exist_ok=True)
        workflow.s3_download('iter-dataset', dp_dataset_dir)
        workflow.s3_download('train-deepmd', dp_models_dir)
        workflow.s3_download('screen-model-devi', model_devi_dir)
        try:
            gen_report(dp_models_dir=dp_models_dir,
                    model_devi_dir=model_devi_dir,
                    max_iters=args.max_iters,
                    output_dir=str(args.output_dir))
        except:
            logger.exception('Failed to generate report')
    return 0


def _get_executor_config(args: DynacatTeslaArgs):
    return {
        'executors': {
            'bohrium': {
                'bohrium': {},
                'apps': {
                    'python': {
                        'resource': {
                            'bohrium': {**args.python_resource, 'job_name':'dynacat_tesla_python'},
                        }
                    },
                    'deepmd': {
                        'resource': {
                            'bohrium': {**args.deepmd_resource, 'job_name':'dynacat_tesla_deepmd'},
                        },
                        'dp_cmd': args.deepmd.cmd,
                        'concurrency': args.deepmd.concurrency,
                        'setup_script': args.deepmd.setup_script,
                    },
                    'lammps': {
                        'resource': {
                            'bohrium': {**args.lammps_resource, 'job_name':'dynacat_tesla_lammps'},
                        },
                        'lammps_cmd': args.lammps.cmd,
                        'concurrency': args.lammps.concurrency,
                        'setup_script': args.lammps.setup_script,
                    },
                    'cp2k': {
                        'resource': {
                            'bohrium': {**args.cp2k_resource, 'job_name': 'dynacat_tesla_cp2k'},
                        },
                        'cp2k_cmd': args.cp2k.cmd,
                        'concurrency': args.cp2k.concurrency,
                        'setup_script':  args.cp2k.setup_script,
                    }
                }
            }
        },
        'orchestration': {
            'deepmd': 'bohrium',
            'lammps': 'bohrium',
            'model_devi': 'bohrium',
            'cp2k': 'bohrium'
        },
    }


def _get_workflow_config(args: DynacatTeslaArgs, dp_dataset_config: dict):
    # process system file
    explore_data_file = args.lammps_system_file.get_full_path()
    explore_data_key = os.path.basename(explore_data_file)
    atoms = ase.io.read(explore_data_file, index=0)
    type_map, mass_map = ai2cat.get_type_map(atoms)  # type: ignore

    if args.deepmd_input_template:
        deepmd_template_file = args.deepmd_input_template.get_full_path()
    else:
        deepmd_template_file = str(get_res_path() / 'dynacat' / 'deepmd.json')
    deepmd_template = load_json(deepmd_template_file)
    cp2k_input_template = load_text(args.cp2k_input_template.get_full_path())

    if args.deepmd.steps and args.deepmd.steps > 0:
        deepmd_template['training']['numb_steps'] = args.deepmd.steps
    assert deepmd_template['training']['numb_steps'] > 0, 'numb_steps is required in the input template or input'
    deepmd_template['training']['disp_file'] = 'lcurve.out'  # force to use lcurve.out

    product_vars, broadcast_vars = _get_lammps_vars(args.lammps.explore_vars)

    return {
        'datasets': {
            **dp_dataset_config,
            explore_data_key: {  # data key must be a normal file name or else ase cannot detect the format
                'url': explore_data_file,
            },
        },
        'workflow': {
            'general': {
                'type_map': type_map,
                'mass_map': mass_map,
            },
            'train': {
                'deepmd': {
                    'init_dataset': list(dp_dataset_config.keys()),
                    'input_template': deepmd_template,
                    'compress_model': args.deepmd.compress_model,
                }
            },
            'explore':{
                'lammps': {
                    'systems': [explore_data_key],
                    'nsteps': args.lammps.nsteps,
                    'ensemble': args.lammps.ensemble.value,
                    'timestep': args.lammps.timestep,
                    'sample_freq': args.lammps.sample_freq,
                    'no_pbc': args.lammps.no_pbc,
                    'plumed_config': str_or_none(args.lammps.plumed_config),
                    'product_vars': product_vars,
                    'broadcast_vars': broadcast_vars,
                    'template_vars': dict((item.key, item.value) for item in args.lammps.template_vars),
                }
            },
            'screen':{
                'model_devi': {
                    'metric': args.model_deviation.metric,
                    'decent_range': [args.model_deviation.lo, args.model_deviation.hi],
                }
            },
            'label': {
                'cp2k': {
                    'input_template': cp2k_input_template,
                    'limit': args.cp2k.limit,
                }
            },
        }
    }


def _get_lammps_vars(explore_vars: List[ExploreItem]):
    broadcast_vars = {}
    product_vars = {}
    for item in explore_vars:
        if item.broadcast:
            broadcast_vars[item.key] = parse_string_array(item.value, dtype=float, delimiter=',')
        else:
            product_vars[item.key] = parse_string_array(item.value, dtype=float, delimiter=',')
    return product_vars, broadcast_vars


def _unpack_dpdata(file: str, extract_dir: str):
    extract_dir = os.path.normpath(extract_dir)
    os.makedirs(extract_dir, exist_ok=True)
    shutil.unpack_archive(file, extract_dir=extract_dir)
    # use type.raw to locate the dpdata folder
    paths = glob.glob(f'{extract_dir}/**/type.raw', recursive=True)
    return [ os.path.dirname(p) for p in paths]


def main():
    to_runner(
        DynacatTeslaArgs,
        launch_app,
        version="0.1.0",
    )(sys.argv[1:])


if __name__ == "__main__":
    fire.Fire({
        'main': main,
        'gen_report': gen_report,
    })



from dp.launching.typing import BaseModel, Field, OutputDirectory, InputFilePath, Optional
from dp.launching.typing import Int, String, Enum, Float, Boolean, Set
from dp.launching.cli import to_runner

from dflow_galaxy.app.common import DFlowOptionsMixin, setup_dflow_context
from dflow_galaxy.res import get_cp2k_data_dir
from dflow_galaxy.core.log import get_logger
from ai2_kit.feat import catalysis as ai2cat
from ai2_kit.tool.ase import AseTool
from ai2_kit.tool.dpdata import DpdataTool

from pathlib import Path
import shutil
import sys
import os

from .dflow import run_cp2k_workflow

logger = get_logger(__name__)


def get_cp2k_data_file(name: str):
    data_file =  get_cp2k_data_dir() / name
    return str(data_file)


class SystemTypeOptions(String, Enum):
    metal = "metal"
    semi = 'semi'


class AccuracyOptions(String, Enum):
    low = "low"
    medium = "medium"
    high = "high"


class BasicSetOptions(String, Enum):
    HFX_BASIS = "HFX_BASIS"
    BASIS_ADMM = "BASIS_ADMM"
    BASIS_def2_QZVP_RI_ALL = "BASIS_def2_QZVP_RI_ALL"
    BASIS_MOLOPT_LnPP1 = "BASIS_MOLOPT_LnPP1"
    BASIS_MOLOPT_AcPP1 = "BASIS_MOLOPT_AcPP1"
    BASIS_SET = "BASIS_SET"
    BASIS_MOLOPT_UZH = "BASIS_MOLOPT_UZH"
    BASIS_ADMM_ae = "BASIS_ADMM_ae"
    ALL_BASIS_SETS = "ALL_BASIS_SETS"
    BASIS_LRIGPW_AUXMOLOPT = "BASIS_LRIGPW_AUXMOLOPT"
    BASIS_MOLOPT = "BASIS_MOLOPT"
    BASIS_MINIX = "BASIS_MINIX"
    BASIS_MOLOPT_UCL = "BASIS_MOLOPT_UCL"
    BASIS_ZIJLSTRA = "BASIS_ZIJLSTRA"
    GTH_BASIS_SETS = "GTH_BASIS_SETS"
    EMSL_BASIS_SETS = "EMSL_BASIS_SETS"
    BASIS_MINBAS = "BASIS_MINBAS"
    BASIS_ccGRB_UZH = "BASIS_ccGRB_UZH"
    BASIS_ADMM_MOLOPT = "BASIS_ADMM_MOLOPT"
    BASIS_ADMM_UZH = "BASIS_ADMM_UZH"
    BASIS_pob = "BASIS_pob"
    BASIS_PERIODIC_GW = "BASIS_PERIODIC_GW"
    BASIS_RI_cc_TZ = "BASIS_RI_cc-TZ"
    BASIS_MOLOPT_LnPP2 = "BASIS_MOLOPT_LnPP2"


class PotentialOptions(String, Enum):
    GTH_SOC_POTENTIALS = "GTH_SOC_POTENTIALS"
    LnPP2_POTENTIALS = "LnPP2_POTENTIALS"
    HF_POTENTIALS = "HF_POTENTIALS"
    GTH_POTENTIALS = "GTH_POTENTIALS"
    ALL_POTENTIALS = "ALL_POTENTIALS"
    ECP_POTENTIALS_pob_TZVP_rev2 = "ECP_POTENTIALS_pob-TZVP-rev2"
    ECP_POTENTIALS = "ECP_POTENTIALS"
    NLCC_POTENTIALS = "NLCC_POTENTIALS"
    AcPP1_POTENTIALS = "AcPP1_POTENTIALS"
    LnPP1_POTENTIALS = "LnPP1_POTENTIALS"


class Cp2kLightningArgs(BaseModel, DFlowOptionsMixin):

    dry_run: Boolean = Field(
        default = True,
        description="Generate configuration file without running the simulation")

    system_file: InputFilePath = Field(
        description="A system file as the initial structure of AIMD simulation, can be xyz, cif, POSCAR format")

    system_type: SystemTypeOptions = Field(
        default=SystemTypeOptions.metal,
        description="Type of the system, semi for semi-conductor, metal for metal")

    accuracy: AccuracyOptions = Field(
        default=AccuracyOptions.medium,
        description="Accuracy of the simulation, the higher the accuracy, the longer the simulation time")

    temperature: Float = Field(
        default=300.0,
        description="Temperature of the simulation in K")

    steps: Int = Field(
        default=1000,
        description="Number of steps in the simulation")

    timestep: Float = Field(
        default=0.5,
        description="Time step of the simulation in fs")

    basis_set: Set[BasicSetOptions] = Field(
        default=[BasicSetOptions.BASIS_MOLOPT],
        min_items=1,
        description='Select the basis set for the simulation')

    potential: PotentialOptions = Field(
        default=PotentialOptions.GTH_POTENTIALS,
        description='Select the potential for the simulation')

    output_dir: OutputDirectory = Field(
        default="./output",
        description="Output directory for the simulation results")

    cp2k_image: String = Field(
        default='registry.dp.tech/dptech/cp2k:11',
        description="Docker image for running CP2K simulation")

    cp2k_device_model: String = Field(
        default='c32_m64_cpu',
        description="Device model for CP2K simulation")

    cp2k_script: String = Field(
        default= '\n'.join([
            '# 1. the output file must be named as `output`',
            '# 2. cp2k.aimd.inp for aimd and cp2k.dft.inp for dft',
            '# Note that different container may have different setup',
            'source /opt/cp2k-toolchain/install/setup',
            'mpirun -n 32 cp2k.popt -i cp2k.aimd.inp &> output',
            'rm *.wfn  # reduce the size of output file',
        ]),
        format='multi-line',
        description="Script to run CP2K simulation, note that it depends on the docker image")


def launch_app(args: Cp2kLightningArgs) -> int:
    # stage 1: generate cp2k input file
    basis_set_files = [v.value for v in args.basis_set]
    assert len(basis_set_files) > 0, 'basis_set must be set'
    potential_file = args.potential.value
    for f in basis_set_files + [potential_file]:
        f = get_cp2k_data_file(f)
        # copy data file to cwd
        # don't use absolute path as the config file will be use in docker
        shutil.copy(f, '.')

    system_file = args.system_file.get_full_path()

    # create a closure to generate cp2k input file
    def _gen_cp2k_input(out_dir: str, aimd: bool):
        config_builder = ai2cat.ConfigBuilder()
        config_builder.load_system(system_file).gen_cp2k_input(
            out_dir=out_dir,
            aimd=aimd,
            # common options
            style=args.system_type.value,
            temp=args.temperature,
            steps=args.steps,
            timestep=args.timestep,
            basis_set_file=basis_set_files,
            potential_file=potential_file,
            parameter_file='dftd3.dat',
        )
    _gen_cp2k_input('aimd', aimd=True)
    _gen_cp2k_input('dft', aimd=False)

    out_dir = Path(args.output_dir)
    shutil.move('aimd/cp2k.inp', out_dir / 'cp2k.aimd.inp')
    shutil.move('dft/cp2k.inp', out_dir / 'cp2k.dft.inp')
    shutil.move('aimd/coord_n_cell.inc', out_dir / 'coord_n_cell.inc')


    if args.dry_run:
        logger.info('skip dflow run due to dry_run is set to True')
        return 0

    # stage 2: run cp2k with dflow
    setup_dflow_context(args)
    cp2k_output_dir = run_cp2k_workflow(
        input_dir=str(args.output_dir),
        out_dir=str(args.output_dir),
        cp2k_device_model=str(args.cp2k_device_model),
        cp2k_image=str(args.cp2k_image),
        cp2k_script=str(args.cp2k_script),
    )

    # stage 3: post-process cp2k output
    # convert cp2k output to xyz file and dpdata set
    AseTool().read(
        os.path.join(cp2k_output_dir, '*-pos-1.xyz')
    ).set_by_ref(
        system_file
    ).write(str(out_dir / 'aimd.xyz' ))

    DpdataTool().read(
        cp2k_output_dir, fmt='cp2kdata/md', cp2k_output_name='output'
    ).write(str(out_dir / 'dp-dataset'))

    return 0

def main():
    to_runner(
        Cp2kLightningArgs,
        launch_app,
        version="0.1.0",
    )(sys.argv[1:])


if __name__ == "__main__":
    main()



from typing import Optional, TypeVar
from ai2_kit.core.util import list_split
import sys
import os


from .types import ListStr, SliceIndex

T = TypeVar('T')

def ensure_dir(path: str):
    os.makedirs(path, exist_ok=True)


def ensure_dirname(path: str):
    os.makedirs(os.path.dirname(path), exist_ok=True)


def not_none(v: Optional[T], msg: str = '') -> T:
    if v is None:
        raise ValueError(msg)
    return v


def ensure_str(s: ListStr):
    if isinstance(s, list):
        return '\n'.join(s)
    return s


def bash_ln_cmd(from_path: str, to_path: str):
    """
    The reason to `rm -d` to_path is to workaround the limit of ln.
    `ln` command cannot override existed directory,
    so we need to ensure to_path is not existed.
    Here we use -d option instead of -rf to avoid remove directory with content.
    The error of `rm -d` is suppressed as it will fail when to_path is file.
    `-T` option of `ln` is used to avoid some unexpected result.
    """
    return f'rm -d {to_path} || true && ln -sfT {from_path} {to_path}'


def safe_ln(from_path: str, to_path: str, method=None):
    if method is None:
        method = os.system
    method(bash_ln_cmd(from_path, to_path))


def select_chunk(in_list: list, n: int, i: int):
    assert 0 <= i < n, f'nth should be in range [0, {n})'
    return list_split(sorted(in_list), n)[i]


def inspect_dir(dir: str):
    os.system(bash_inspect_dir(dir))


def bash_inspect_dir(dir: str):
    return f"""echo "Show content of {dir}" && find {dir}"""


def bash_iter_ls_slice(search_pattern: str, /, n: int, i: SliceIndex, script: ListStr, opt: str = '',
                       it_var='ITEM', python_cmd: str = 'python'):
    """
    Generate a bash snippet to slice the result of `ls` command,
    and iterate over the selected chunk.

    :param search_pattern: search pattern for directories
    :param n: number of chunks
    :param i: chunk index
    :param script: bash script to process each directory
    :param it_var: variable name for each item
    """
    return '\n'.join([
        f'_LS_RESULT=$(ls -1 {opt} {search_pattern} | sort)',
        bash_slice(in_var='_LS_RESULT', n=n, i=i, out_var='_LS_CHUNK', python_cmd=python_cmd),
        bash_iter_var(in_var='_LS_CHUNK', script=script, it_var=it_var),
    ])


def bash_iter_var(in_var: str, script: ListStr, it_var='ITEM', report_progress=True):
    """
    Generate a bash snippet to iterate over lines of a variable

    :param in_var: variable name of input data
    :param script: bash script to process each line
    """
    script = ensure_str(script)

    argo_progress_init, argo_progress_count  = '', ''
    if report_progress:
        argo_progress_init = f'_ARGO_I=0 && echo "$_ARGO_I/$_N_LINES" > $ARGO_PROGRESS_FILE'
        argo_progress_count = '_ARGO_I=$((_ARGO_I + 1)) && echo "$_ARGO_I/$_N_LINES" > $ARGO_PROGRESS_FILE'

    return f"""_N_LINES=$(grep . <<< "${in_var}" | wc -l)
{argo_progress_init}
if [ $_N_LINES -ne 0 ]; then
while IFS= read -r {it_var}; do
{script}
{argo_progress_count}
done <<< "${in_var}"
fi"""



def bash_slice(in_var: str, n: int, i: SliceIndex, out_var: str,
               python_cmd: str = 'python'):
    """
    Generate a bash snippet to slice a multi-line string variable
    into n chunks and select the ith chunk

    :param in_var: variable name of input multi-line string
    :param n: number of chunks
    :param i: chunk index
    :param out_var: variable name to store the selected chunk
    """
    return f"""# bash_slice({in_var}, {n}, {i}, {out_var})
which python3 && PY_CMD=python3 || PY_CMD=python  # prefer python3
which "{python_cmd}" && PY_CMD="{python_cmd}" || true  # prefer user specified python
{out_var}=$(_IN_DATA="${in_var}" _SLICE_N={n} _SLICE_I={i} $PY_CMD << EOF
import sys,os
lines = os.environ['_IN_DATA'].split('\\n')
n = int(os.environ['_SLICE_N'])
i = int(os.environ['_SLICE_I'])
lines = [line for line in lines if line.strip()]
chunk_size = max(1, len(lines) // n)

start = i * chunk_size
end = (i + 1) * chunk_size if i < n - 1 else len(lines)
sys.stdout.write('\\n'.join(lines[start:end]))
EOF
)
# bash_slice end"""


def yes_or_no(msg: str, default: bool = False):
    """
    prompt user for input, if user press n or N, return False
    if user press y or Y, return True
    otherwise, ask again
    """
    msg = f'{msg} [Y/n]: ' if default else f'{msg} [n/Y]: '

    while True:
        ans = input(msg)
        if not ans:
            return default
        if ans.lower() == 'y':
            return True
        if ans.lower() == 'n':
            return False
        print('Invalid input, please input y or n')


def resolve_ln(path: str, mv=False):
    """
    This command will iterate over the files in the given directory recursively,
    if the file is a soft link, it will resolve the link and
    copy the resolved file to the same directory with the same name.

    :param path: the directory to resolve symlinks
    :param mv: if True, move the resolved file to the same directory with the same name
    """
    if os.path.islink(path):
        resolved_path = os.path.realpath(path)
        resolved_file = os.path.join(os.path.dirname(path), os.path.basename(resolved_path))
        cmd = 'mv' if mv else 'cp -f'
        os.system(f'rm {resolved_file} && {cmd} {resolved_path} {resolved_file}')
        # log is too verbose
        # print(f'replace {path} with {resolved_path}', file=sys.stderr)

    elif os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                resolve_ln(file_path, mv=mv)
            for dir in dirs:
                dir_path = os.path.join(root, dir)
                resolve_ln(dir_path, mv=mv)


def parse_string_array(s: str, dtype=None, delimiter=None):
    arr = [x.strip() for x in s.split(delimiter)]
    if dtype:
        arr = [dtype(x) for x in arr]
    return arr

def str_or_none(s):
    """
    return None if s is empty string
    """
    if not s:
        return None
    return s



from typing import List, Literal
from ai2_kit.core.artifact import Artifact, ArtifactDict
import glob
from dflow_galaxy.core.util import yes_or_no


LabelApp = Literal['cp2k', 'vasp', 'abacus', 'dpa2']
TrainApp = Literal['deepmd']
ExploreApp = Literal['lammps']


def resolve_artifact(a: Artifact) -> List[ArtifactDict]:
    a_dict = a.to_dict()

    if a.includes is not None:
        urls = glob.glob(f'{a.url}/{a.includes}')
        return [{**a_dict, 'url': url, 'includes': None } for url in urls ]
    return [a_dict]


def resolve_artifacts(a_list: List[Artifact]) -> List[ArtifactDict]:
    result = []
    for a in a_list:
        result.extend(resolve_artifact(a))
    return result


class StepSwitch:
    def __init__(self, enable: bool):
        """
        Control whether to skip a step.

        If enable is False, the step will never be skipped

        Otherwise, the step will be skipped by default,
        unless user answer yes to not skip it.
        And the answer will be remembered so that the following steps will
        """
        self._skip = enable

    def shall_skip(self, step: str):
        if self._skip:
            self._skip = yes_or_no(f'Do you want to skip this step: {step}?', default=True)
        return self._skip



from typing import TypeVar, Annotated, Union, List, Literal
from enum import IntEnum, auto

T = TypeVar('T')
class Symbol(IntEnum):
    INPUT_PARAMETER = auto()
    INPUT_ARTIFACT = auto()
    OUTPUT_PARAMETER = auto()
    OUTPUT_ARTIFACT = auto()


InputParam = Annotated[T, Symbol.INPUT_PARAMETER]
InputArtifact = Annotated[str, Symbol.INPUT_ARTIFACT]
OutputParam = Annotated[T, Symbol.OUTPUT_PARAMETER]
OutputArtifact = Annotated[str, Symbol.OUTPUT_ARTIFACT]


ListStr = Union[List[str], str]

SliceIndex = Union[int, Literal['{{item}}']]


from typing import List, Literal, Tuple
from collections import namedtuple
from dataclasses import dataclass
from pathlib import Path
from itertools import groupby
import glob
import os
import re

from joblib import Parallel, delayed
from tabulate import tabulate
import pandas as pd
import ase.io

from dflow_galaxy.core.pydantic import BaseModel

from dflow_galaxy.core.dispatcher import PythonApp, create_dispatcher, ExecutorConfig
from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core import types
from dflow_galaxy.core.util import inspect_dir

from ai2_kit.core.util import load_text, dump_text

from .lib import ExploreApp


class ModelDeviConfig(BaseModel):
    metric: Literal["max_devi_v",  "min_devi_v",  "avg_devi_v",  "max_devi_f",  "min_devi_f",  "avg_devi_f"] = 'max_devi_f'
    decent_range: Tuple[float, float]


@dataclass(frozen=True)
class RunModelDeviTasksArgs:
    explore_dir: types.InputArtifact
    persist_dir: types.OutputArtifact

ModelDeviResult = namedtuple('_ModelDeviResult', ['data_dir', 'decent_xyz', 'ancestor', 'total', 'n_good', 'n_decent', 'n_poor'])

class RunModelDeviTasksFn:

    def __init__(self, config: ModelDeviConfig, workers: int, type_map: List[str], explore_app: ExploreApp):
        self.config = config
        self.workers = workers
        self.type_map = type_map
        self.explore_app = explore_app

    def __call__(self, args: RunModelDeviTasksArgs):
        inspect_dir(args.explore_dir)
        persis_dir = Path(args.persist_dir)
        persis_dir.mkdir(exist_ok=True)
        if self.explore_app == 'lammps':
            data_dirs = sorted(glob.glob(f'{args.explore_dir}/tasks/*/persist'))
            results: List[ModelDeviResult] = Parallel(n_jobs=self.workers)(
                delayed(self._process_lammps_dir)(Path(d)) for d in data_dirs
            ) # type: ignore
        else:
            raise ValueError(f'unsupported explore app: {self.explore_app}')

        # generate report
        headers = ['src', 'total', 'good', 'decent', 'poor', 'good%', 'decent%', 'poor%']
        rows = [ ]
        _pp = lambda a, b: f'{a / b * 100:.2f}%'
        for r in results:
            row = [os.path.relpath(r.data_dir, args.explore_dir), r.total, r.n_good, r.n_decent, r.n_poor,
                   _pp(r.n_good, r.total), _pp(r.n_decent, r.total), _pp(r.n_poor, r.total)]
            rows.append(row)

        report_text = tabulate(rows, headers=headers, tablefmt='tsv')
        dump_text(report_text, str(persis_dir / 'report.tsv'))

        # dump decent xyz
        results = sorted(results, key=lambda r: r.ancestor)
        for ancestor, group in groupby(results, key=lambda r: r.ancestor):
            assert ancestor, f'ancestor should not be empty'
            group = list(group)
            assert group, f'group should not be empty'
            decent_xyz_files = [str(r.decent_xyz) for r in group if r.decent_xyz is not None]
            if not decent_xyz_files:
                print(f'no decent files for ancestor {ancestor}')
                continue

            result_dir: Path = persis_dir / 'system' / ancestor
            result_dir.mkdir(parents=True, exist_ok=True)

            # merge xyz files, use cat for the sake of performance
            os.system(f'cat {" ".join(decent_xyz_files)} > {result_dir / "decent.xyz"}')
            dump_text(ancestor, str(result_dir / 'ANCESTOR'))

    def _process_lammps_dir(self, data_dir: Path):
        col = self.config.metric

        model_devi_file = data_dir / 'model_devi.out'
        traj_dir = data_dir / 'traj'

        with open(model_devi_file, 'r') as f:
            f.seek(1)  # skip the leading '#'
            df = pd.read_csv(f, delim_whitespace=True)
        lo, hi = self.config.decent_range

        good_df = df[df[col] < lo]
        decent_df = df[(df[col] >= lo) & (df[col] < hi)]
        poor_df = df[df[col] > hi]

        traj_files = glob.glob(f'{traj_dir}/*.lammpstrj')
        assert traj_files, f'no traj files is found in {traj_dir}'

        # pick decent files
        traj_files = sorted(traj_files, key=get_lammpstrj_frame_no)  # align
        decent_files = [traj_files[i] for i in decent_df.index]

        # merge decent files
        decent_xyz = None
        if len(decent_df) > 0:
            decent_trj = data_dir / 'decent.lammpstrj'
            decent_xyz = data_dir / 'decent.xyz'
            # use cat for the sake of performance
            os.system(f'cat {" ".join(decent_files)} > {decent_trj}')
            self._dump_lammpstrj_to_xyz(decent_trj, decent_xyz)

        return ModelDeviResult(
            data_dir=data_dir,
            decent_xyz=decent_xyz,
            ancestor=load_text(data_dir / 'ANCESTOR'),
            total=len(df),
            n_good=len(good_df),
            n_decent=len(decent_df),
            n_poor=len(poor_df),
        )

    def _dump_lammpstrj_to_xyz(self, lmptrj_file: Path, xyz_file: Path):
        atoms_list = ase.io.read(lmptrj_file, ':', format='lammps-dump-text', specorder=self.type_map)
        ase.io.write(xyz_file, atoms_list, format='extxyz')


def get_lammpstrj_frame_no(filename):
    filename = os.path.basename(filename)
    m = re.match(r'^(\d+)', filename)
    assert m, f'unexpected lammpstrj file name: {filename}'
    return int(m.group(1))


def provision_model_devi(builder: DFlowBuilder, ns: str, /,
                         config: ModelDeviConfig,
                         executor: ExecutorConfig,
                         python_app: PythonApp,
                         type_map: List[str],

                         explore_app: ExploreApp,
                         explore_data_url: str,
                         persist_data_url: str,
                         ):
    run_tasks_fn = RunModelDeviTasksFn(config, workers=python_app.max_worker,
                                       type_map=type_map, explore_app=explore_app)
    run_tasks_step = builder.make_python_step(run_tasks_fn, uid=f'{ns}-run-task',
                                              setup_script=python_app.setup_script,
                                              executor=create_dispatcher(executor, python_app.resource))(
        RunModelDeviTasksArgs(
            explore_dir=explore_data_url,
            persist_dir=persist_data_url,
        )
    )
    builder.add_step(run_tasks_step)



from dp.launching.report import Report, ReportSection, ChartReportElement
from typing import List
import glob
import os

from dflow_galaxy.core.util import parse_string_array

def gen_report(dp_models_dir: str,
               model_devi_dir: str,
               max_iters: int,
               output_dir: str):
    # build report sections, each iter per section, from last to first
    os.makedirs(output_dir, exist_ok=True)
    sections = []
    for i in reversed(range(max_iters)):
        iter_str = f'iter/{i:03d}'
        lcurve_files = glob.glob(f'{dp_models_dir}/{iter_str}/**/lcurve.out', recursive=True)
        model_devi_files = glob.glob(f'{model_devi_dir}/{iter_str}/**/report.tsv', recursive=True)

        if not (lcurve_files or model_devi_files):
            continue
        sections.append(_gen_report_section(i, lcurve_files, model_devi_files))
    # write report
    report = Report(title='DynaCat TESLA', sections=sections)
    report.save(output_dir)


def _gen_report_section(iter: int, lcurve_files: List[str], model_devi_files: List[str]):
    elements = []
    if lcurve_files:
        for i, f in enumerate(sorted(lcurve_files)):
            name = os.path.dirname(f)
            echart = _gen_lcurve_echart(f)
            element = ChartReportElement(
                title=f'Learning Curve of training: {name}',
                options=echart,
            )
            elements.append(element)

    if model_devi_files:
        # there should be only 1 file in each iteration
        f = model_devi_files[0]
        echart = _gen_model_devi_stats_echart(f)
        element = ChartReportElement(
            title='Model Deviation Statistics',
            options=echart,
        )
        elements.append(element)

    section = ReportSection(
        title=f'Result of Iteration {iter:03d}',
        ncols=2,
        elements=elements,
    )
    return section


def _gen_model_devi_stats_echart(file: str):
    data_dict = _load_model_devi_stats(file)
    echart = {
        'tooltip': {
            'trigger': 'axis',
            'axisPointer': {
                'type': 'shadow'
            }
        },
        'legend': {
            'data': ['Good', 'Decent', 'Poor']
        },
        'grid': {
            'left': '10%',
            'right': '10%',
            'bottom': '3%',
            'containLabel': True
        },
        'xAxis': {
            'type': 'value'
        },
        'yAxis': {
            'type': 'category',
            'data': data_dict['src']
        },
        'series': [
            {
                'name': 'Good',
                'type': 'bar',
                'stack': 'total',
                'label': {
                    'show': False,
                },
                'itemStyle': {
                    'color': '#67C23A'  # Green color to indicate 'good' is better
                },
                'data': [int(d) for d in data_dict['good']]
            },
            {
                'name': 'Decent',
                'type': 'bar',
                'stack': 'total',
                'label': {
                    'show': False
                },
                'itemStyle': {
                    'color': '#E6A23C'  # Orange color to indicate 'decent'
                },
                'data': [int(d) for d in data_dict['decent']]
            },
            {
                'name': 'Poor',
                'type': 'bar',
                'stack': 'total',
                'label': {
                    'show': False
                },
                'itemStyle': {
                    'color': '#F56C6C'  # Red color to indicate 'poor' (danger)
                },
                'data': [int(d) for d in data_dict['poor']]
            }
        ]
    }
    return echart


def _load_model_devi_stats(file: str):
    header = None
    with open(file, newline='') as fp:
        headers = parse_string_array(next(fp), delimiter='\t')
        data_dict = {name: [] for name in headers}
        for line in fp:
            line = line.strip()
            if not line:
                continue
            values = parse_string_array(line, delimiter='\t')
            for i, header in enumerate(headers):
                data_dict[header].append(values[i])
    return data_dict


def _gen_lcurve_echart(file: str):
    data_dict = _load_lcurve(file)
    x = data_dict.pop('step')
    echart = {
        'tooltip': {
            'trigger': 'axis',
        },
        'xAxis': {
            'type': 'category',
            'name': 'Step',
            'data': x,
        },
        'yAxis': [
            {
                'type': 'log',
                'name': 'RMSE',
                'position': 'left',
            },
            {
                'type': 'log',
                'name': 'Learning Rate',
                'position': 'right',
            }
        ],
        'legend': {
            'data': [name for name in data_dict.keys()],
        },
        'series': [],
    }
    for name, data in data_dict.items():
        echart['series'].append({
            'name': name,
            'data': data,
            'type': 'line',
            'smooth': True,
            'yAxisIndex': 0 if name != 'lr' else 1,
        })
    return echart


def _load_lcurve(file: str):
    header = None
    data = []
    with open(file, encoding='utf-8') as fp:
        for line in fp:
            line = line.strip()
            if not line:
                continue
            if line.startswith('#'):
                if header is not None:
                    continue  # ignore comment line
                header = line[1:].split()
            else:
                data.append([float(x) for x in line.split()])
    assert header is not None, 'Failed to parse lcurve file'
    # convert to series
    series = {}
    for i, h in enumerate(header):
        series[h] = [d[i] for d in data]
    return series


import dflow
from dflow.plugins import bohrium
from dflow.plugins.bohrium import TiefblueClient

from typing import Optional
import getpass
import os


def bohrium_nb_setup_dflow(username: Optional[str], password: Optional[str] = None, project_id: Optional[int] = None):
    if username is None:
        username = input("Enter Bohrium Email: ")
    if password is None:
        password = getpass.getpass("Enter Bohrium password: ")
    if project_id is None:
        project_id = int(os.environ.get("PROJECT_ID"))  # type: ignore

    dflow.config["host"] = "https://workflows.deepmodeling.com"
    dflow.config["k8s_api_server"] = "https://workflows.deepmodeling.com"
    bohrium.config["username"] = username
    bohrium.config["password"] = password
    bohrium.config["project_id"] = project_id
    dflow.s3_config["repo_key"] = "oss-bohrium"
    dflow.s3_config["storage_client"] = TiefblueClient()



from dp.launching.typing import BaseModel, Field, OutputDirectory, InputFilePath, Optional, Dict
from dp.launching.typing import Int, String, Enum, Float, Boolean
from dp.launching.cli import to_runner, default_minimal_exception_handler

from dflow_galaxy.app.common import DFlowOptionsMixin, setup_dflow_context, EnsembleOptions
from dflow_galaxy.core.log import get_logger
from ai2_kit.feat import catalysis as ai2cat
from ai2_kit.core.util import dump_text, dump_json


from pathlib import Path
import shutil
import sys
import os

from .dflow import run_lammps_workflow
from .report import gen_report

logger = get_logger(__name__)


class DynaCatMdArgs(BaseModel, DFlowOptionsMixin):

    dry_run: Boolean = Field(
        default = True,
        description="Generate configuration file without running the simulation")

    system_file: InputFilePath = Field(
        description="A system file as the initial structure of LAMMPS simulation, can be xyz, cif, POSCAR, etc.")

    # TODO: support multiple deepmd models
    deepmd_model: InputFilePath = Field(
        description="Deepmd model file for LAMMPS simulation")

    ensemble: EnsembleOptions = Field(
        default=EnsembleOptions.csvr,
        description='Ensemble of LAMMPS simulation')

    temperature: Float = Field(
        default=330,
        description='Temperature of LAMMPS simulation')

    plumed_config: String = Field(
        format='multi-line',
        description='Plumed configuration file for metadynamics simulation')

    pressure: Float = Field(
        default=-1,
        description='Pressure of LAMMPS simulation, should be -1 unless it is NPT ensemble')

    steps: Int = Field(
        default=20000,
        description='Steps of LAMMPS simulation')

    step_size: Float = Field(
        default=0.0005,
        description='Time step size of LAMMPS simulation in ps')

    sample_freq: Int = Field(
        title='Sampling Frequency',
        default=10,
        description='Sampling frequency of LAMMPS simulation')

    extra_args: Dict[String, Float] = Field(
        title='Extra Arguments',
        default={
            'tau_t': 0.1,
            'tau_p': 0.5,
            'time_const': 0.1,
        },
        description="Extra arguments for LAMMPS simulation, e.g. tau_t, tau_p, time_const, etc. Don't remove or add extra arguments if you are not sure about it."
    )

    output_dir : OutputDirectory = Field(
        default='./output',
        description="Output directory of LAMMPS simulation")

    lammps_image: String = Field(
        default='registry.dp.tech/dptech/dpmd:2.2.8-cuda12.0',
        description="Docker image for running LAMMPS simulation")

    lammps_device_model: String = Field(
        default='c8_m32_1 * NVIDIA V100',
        description="Device model for LAMMPS simulation")

    lammps_script: String = Field(
        default= '\n'.join([
            '# LAMMPS input can be referenced as lammps.inp',
            '# Note that different container may have different setup',
            'lmp -in lammps.inp &> lammps.out',
            '# Run plumed to calculate free energy surface',
            'plumed sum_hills --hills HILLS --mintozero --outfile fes.dat'
        ]),
        format='multi-line',
        description="Script to run LAMMPS simulation, note that it depends on the docker image")


def launch_app(args: DynaCatMdArgs) -> int:
    config_builder = ai2cat.ConfigBuilder()

    shutil.copy(args.deepmd_model, 'dp-model.pb')
    if not args.plumed_config:
        raise ValueError('plumed_config is required for metadynamics simulation')

    dump_text(args.plumed_config, 'plumed.inp')

    logger.info(f'type of system_file: {type(args.system_file)}')
    config_builder.load_system(args.system_file.get_full_path()).gen_lammps_input(
        out_dir=args.output_dir,
        nsteps=args.steps,
        temp=args.temperature,
        sample_freq=args.sample_freq,
        pres=args.pressure,
        abs_path=False,
        ensemble=args.ensemble,
        dp_models=['dp-model.pb'],
        **args.extra_args,
    )
    shutil.move('plumed.inp', args.output_dir)
    shutil.move('dp-model.pb', args.output_dir)

    if args.dry_run:
        return 0

    setup_dflow_context(args)
    lammps_output_dir = run_lammps_workflow(
        input_dir=str(args.output_dir),
        out_dir=str(args.output_dir),
        lammps_image=str(args.lammps_image),
        lammps_device_model=str(args.lammps_device_model),
        lammps_script=str(args.lammps_script),
    )

    try:
        gen_report(lammps_output_dir=lammps_output_dir,
                   output_dir=str(args.output_dir))
    except:
        logger.exception('Failed to generate report')
    return 0


def main():
    to_runner(
        DynaCatMdArgs,
        launch_app,
        version="0.1.0",
    )(sys.argv[1:])


if __name__ == '__main__':
    main()



from dataclasses import dataclass
from dflow_galaxy.core import types, dispatcher
from dflow_galaxy.core.dflow_builder import DFlowBuilder
import shutil
import os


@dataclass(frozen=True)
class RunLammpsArgs:
    input_dir : types.InputArtifact
    output_dir: types.OutputArtifact


class RunLammpsFn:
    def __init__(self, lammps_script: str):
        self.lammps_script = lammps_script

    def __call__(self, args: RunLammpsArgs):
        script = [
            f'cd {args.input_dir}',
            self.lammps_script,
            f'mkdir -p {args.output_dir}',
            f'mv * {args.output_dir}',
        ]
        return script


def run_lammps_workflow(input_dir: str,
                        out_dir: str,
                        lammps_image: str,
                        lammps_device_model: str,
                        lammps_script: str):

    # bohrium dispatcher will be configured in bohrium.config
    # so here we just leave it empty
    bohrium_config = dispatcher.BohriumConfig()

    # start to build workflow
    dflow_builder = DFlowBuilder('lammps', s3_prefix='lammps')

    # setup and add lammps step to workflow
    lammps_res = dispatcher.Resource(
        bohrium=dispatcher.BohriumInputData(
            image_name=lammps_image,
            scass_type=lammps_device_model,
            disk_size=100,
            job_name='dynacat_md',
        )
    )

    dflow_builder.s3_upload(input_dir, 'lammps_input')
    lammps_executor = dispatcher.create_bohrium_dispatcher(bohrium_config, lammps_res)
    lammps_fn = RunLammpsFn(lammps_script=lammps_script)
    lammps_step = dflow_builder.make_bash_step(lammps_fn, executor=lammps_executor)(
        RunLammpsArgs(input_dir='s3://./lammps_input', output_dir='s3://./lammps_output')
    )
    dflow_builder.add_step(lammps_step)

    # run workflow
    dflow_builder.run()

    # download artifacts to out_dir
    dflow_builder.s3_download('lammps_output')
    shutil.unpack_archive('lammps_output', out_dir, format='gztar')
    return os.path.join(out_dir, 'output_dir')




from dflow.op_template import ScriptOPTemplate
from dflow.plugins.dispatcher import DispatcherExecutor
import dflow

from typing import Final, Callable, TypeVar, Optional, Union, Generic, Dict, Any, Iterable, get_args, get_origin

from dataclasses import fields, is_dataclass
from urllib.parse import urlparse
from collections import namedtuple
from pathlib import Path
from uuid import uuid4
import cloudpickle as cp
import tempfile
import tarfile
import hashlib
import inspect
import base64
import shutil
import shlex
import bz2
import os


from .util import resolve_ln
from .log import get_logger
from . import types

logger = get_logger(__name__)

T_ARGS = TypeVar('T_ARGS')
T_RESULT = TypeVar('T_RESULT')

DFLOW_ARTIFACT = Union[str, dflow.S3Artifact, dflow.OutputArtifact, dflow.InputArtifact]


class Step(Generic[T_ARGS, T_RESULT]):
    def __init__(self, df_step: dflow.Step):
        self.df_step = df_step

    @property
    def args(self) -> T_ARGS:
        return ObjProxy(self.df_step.inputs.parameters,
                        self.df_step.inputs.artifacts,
                        self.df_step.outputs.artifacts)  # type: ignore

    @property
    def result(self) -> T_RESULT:
        return ObjProxy(self.df_step.outputs.parameters)  # type: ignore

Steps = Union[Step, Iterable['Steps']]


class ObjProxy:
    def __init__(self, *obj):
        self.objs = obj

    def __getattr__(self, name):
        for obj in self.objs:
            if name in obj:
                return obj[name]
        raise AttributeError(f'{name} not found in {self.objs}')


def parse_dflow_field(field):
    name = field.name
    _type = field.type

    args = get_args(_type)
    origin = get_origin(_type)
    optional = False

    if origin is Union:
        if type(None) not in args or len(args) != 2:
            raise ValueError(f'Invalid Union type `{_type}` for field `{name}`, you may want to use Optional[T]?')
        _type = args[0]
        origin = get_origin(_type)
        args = get_args(_type)
        optional = True

    error_msg = f'Invalid type `{_type}` for field `{name}`, must be one of InputParam, InputArtifact, OutputParam, OutputArtifact, Annotated[str, dflow.InputArtifact(...)], Annotated[str, dflow.OutputArtifact(...)]'
    metadata = getattr(_type, '__metadata__', None)
    if metadata is None or len(metadata) != 1:
        raise ValueError(error_msg)

    if metadata[0] not in (
        types.Symbol.INPUT_PARAMETER,
        types.Symbol.INPUT_ARTIFACT,
        types.Symbol.OUTPUT_ARTIFACT,
    ) and not isinstance(metadata[0], dflow.InputArtifact) and not isinstance(metadata[0], dflow.OutputArtifact):
        raise ValueError(error_msg)

    return _type, optional


def pickle_converts(obj, pickle_module='cp', bz2_module='bz2', base64_module='base64'):
    """
    convert an object to its pickle string form
    """
    obj_pkl = cp.dumps(obj, protocol=cp.DEFAULT_PROTOCOL)
    compress_level = 5 if len(obj_pkl) > 4096 else 1
    compressed = bz2.compress(obj_pkl, compress_level)
    obj_b64 = base64.b64encode(compressed).decode('ascii')
    return f'{pickle_module}.loads({bz2_module}.decompress({base64_module}.b64decode({repr(obj_b64)})))'


_ParsedField = namedtuple('_ParseField', ['name', 'type', 'optional', 'value'])

def iter_python_step_args(obj):
    """
    Iterate over the input fields of a python step.
    A python step input should be:
    1. A frozen dataclass.
    2. All fields are annotated with InputParam, InputArtifact or OutputArtifact.
    """
    assert is_dataclass(obj), f'{obj} is not a dataclass'
    for f in fields(obj):
        _type, optional = parse_dflow_field(f)
        yield _ParsedField(name=f.name, type=_type,
                           optional=optional, value=getattr(obj, f.name, None))


def iter_python_step_return(obj):
    """
    Iterate over the output fields of a python step.
    A python step output should be:
    1. A dataclass.
    2. All fields are annotated with OutputParam.
    """
    if obj is inspect.Signature.empty or obj is None:
        return

    assert is_dataclass(obj), f'{obj} is not a dataclass'
    for f in fields(obj):
        msg = f'{f.name} is not annotated with OutputParam'
        assert hasattr(f.type, '__metadata__'), msg
        assert f.type.__metadata__ [0] == types.Symbol.OUTPUT_PARAMETER, msg
        yield f, getattr(obj, f.name, None)


_BashTemplate = namedtuple('_BashStep', ['source',
                                         'dflow_input_parameters',
                                         'dflow_input_artifacts',
                                         'dflow_output_artifacts',
                                         ])


def bash_build_template(py_fn: Callable,
                        base_dir: str,
                        setup_script: str = '',
                        default_archive: Optional[str] = 'default',
                        eof: str = '__EOF__') -> _BashTemplate:
    """
    build bash step from a python function
    """
    sig = inspect.signature(py_fn)
    assert len(sig.parameters) == 1, f'{py_fn} should have only one parameter'
    args_type = sig.parameters[next(iter(sig.parameters))].annotation

    input_artifacts_dir = os.path.join(base_dir, 'input-artifacts')
    output_artifacts_dir = os.path.join(base_dir, 'output-artifacts')

    dflow_input_parameters: Dict[str, dflow.InputParameter] = {}
    dflow_input_artifacts: Dict[str, dflow.InputArtifact] = {}
    dflow_output_artifacts: Dict[str, dflow.OutputArtifact] = {}

    args_dict = {}

    source = [
        '#!/bin/bash',
        'set -e',
        'ARGO_PROGRESS_FILE=${ARGO_PROGRESS_FILE:-./argo_progress_file.txt}',
        setup_script,
        ''
        f'mkdir -p {shlex.quote(output_artifacts_dir)}',
        '',
        '# Setup Variables',
    ]

    for f in iter_python_step_args(args_type):
        meta = f.type.__metadata__[0]
        if meta == types.Symbol.INPUT_PARAMETER:
            # input parameter can be a multiline string
            bash_name = f'_DF_INPUT_PARAMETER_{f.name}_'
            source.extend([
                f"{bash_name}=$(cat << {eof}",
                f"{{{{inputs.parameters.{f.name}}}}}",
                eof,
                ')',
            ])
            dflow_input_parameters[f.name] = dflow.InputParameter(name=f.name)
            args_dict[f.name] = f'${bash_name}'

        elif meta == types.Symbol.INPUT_ARTIFACT or isinstance(meta, dflow.InputArtifact):
            bash_name = f'_DF_INPUT_ARTIFACT_{f.name}_'
            path = os.path.join(input_artifacts_dir, f.name)
            source.append(f'{bash_name}={shlex.quote(path)}')
            artifact = meta
            if not isinstance(artifact, dflow.InputArtifact):
                artifact = dflow.InputArtifact(path=path, archive=default_archive, optional=f.optional)  # type: ignore
            dflow_input_artifacts[f.name] = artifact
            args_dict[f.name] = f'${bash_name}'

        elif meta == types.Symbol.OUTPUT_ARTIFACT or isinstance(meta, dflow.OutputArtifact):
            bash_name = f'_DF_OUTPUT_ARTIFACT_{f.name}_'
            path = os.path.join(output_artifacts_dir, f.name)
            source.append(f'{bash_name}={shlex.quote(path)}')
            artifact = meta
            if not isinstance(artifact, dflow.OutputArtifact):
                artifact = dflow.OutputArtifact(path=path, archive=default_archive, optional=f.optional)  # type: ignore
            dflow_output_artifacts[f.name] = artifact

            args_dict[f.name] = f'${bash_name}'
        else:
            raise ValueError(f'unsupported type {f.type}')

    bash_script = py_fn(ObjProxy(args_dict))
    if isinstance(bash_script, list):
        bash_script = '\n'.join(bash_script)
    assert isinstance(bash_script, str), f'{py_fn} should return a string or a list of string'

    source.extend([
        '',
        '#' * 80,
        '',
        bash_script,
    ])

    _dflow_script_check(source, base_dir)
    return _BashTemplate(source='\n'.join(source),
                         dflow_input_parameters=dflow_input_parameters,
                         dflow_input_artifacts=dflow_input_artifacts,
                         dflow_output_artifacts=dflow_output_artifacts)


_PythonTemplate = namedtuple('_PythonStep', ['source', 'fn_str', 'script_path', 'pkg_dir',
                                             'dflow_input_parameters',
                                             'dflow_input_artifacts',
                                             'dflow_output_parameters',
                                             'dflow_output_artifacts',
                                             ])


def python_build_template(py_fn: Callable,
                          base_dir: str,
                          setup_script: str = '',
                          python_cmd: str = 'python3',
                          default_archive: Optional[str] = 'default',
                          eof: str = '__EOF__') -> _PythonTemplate:
    """
    build python template from a python function
    """
    sig = inspect.signature(py_fn)
    assert len(sig.parameters) == 1, f'{py_fn} should have only one parameter'
    args_type = sig.parameters[next(iter(sig.parameters))].annotation
    return_type  = sig.return_annotation

    fn_dir = os.path.join(base_dir, 'python/fn')
    pkg_dir = os.path.join(base_dir, 'python/pkg')
    args_file = os.path.join(fn_dir, 'args.json')
    script_path = os.path.join(fn_dir, 'script.py')
    output_parameters_dir = os.path.join(base_dir, 'output-parameters')
    input_artifacts_dir = os.path.join(base_dir, 'input-artifacts')
    output_artifacts_dir = os.path.join(base_dir, 'output-artifacts')

    dflow_input_parameters: Dict[str, dflow.InputParameter] = {}
    dflow_input_artifacts: Dict[str, dflow.InputArtifact] = {}
    dflow_output_artifacts: Dict[str, dflow.OutputArtifact] = {}
    dflow_output_parameters: Dict[str, dflow.OutputParameter] = {}

    source = [
        '#!/bin/bash',
        'set -e',
        'ARGO_PROGRESS_FILE=${ARGO_PROGRESS_FILE:-./argo_progress_file.txt}',
        '',
        setup_script,
        '',
        f'{python_cmd}  << {eof}',
        'import os, json, tarfile',
        f'base_dir = {repr(base_dir)}',
        f'fn_dir = {repr(fn_dir)}',
        f'pkg_dir = {repr(pkg_dir)}',
        f'args_file = {repr(args_file)}',
        f'script_path = {repr(script_path)}',
        f'output_parameters_dir = {repr(output_parameters_dir)}',
        f'output_artifacts_dir = {repr(output_artifacts_dir)}',
        'os.makedirs(fn_dir, exist_ok=True)',
        'os.makedirs(pkg_dir, exist_ok=True)',
        'os.makedirs(output_parameters_dir, exist_ok=True)',
        'os.makedirs(output_artifacts_dir, exist_ok=True)',
        'args = dict()',
    ]

    for f in iter_python_step_args(args_type):
        meta = f.type.__metadata__[0]
        if meta == types.Symbol.INPUT_PARAMETER:
            # FIXME: may have error in some corner cases
            if _is_str_type(f.type.__origin__):
                val = f'"""{{{{inputs.parameters.{f.name}}}}}"""'
            else:
                val = f'json.loads("""{{{{inputs.parameters.{f.name}}}}}""")'
            dflow_input_parameters[f.name] = dflow.InputParameter(name=f.name)
            source.append(f'args[{repr(f.name)}] = {val}')
        elif meta == types.Symbol.INPUT_ARTIFACT or isinstance(meta, dflow.InputArtifact):
            path = os.path.join(input_artifacts_dir, f.name)
            source.append(f'args[{repr(f.name)}] = {repr(path)}')
            artifact = meta
            if not isinstance(artifact, dflow.InputArtifact):
                artifact = dflow.InputArtifact(path=path, archive=default_archive)  # type: ignore
            dflow_input_artifacts[f.name] = artifact
        elif meta == types.Symbol.OUTPUT_ARTIFACT or isinstance(meta, dflow.OutputArtifact):
            path = os.path.join(output_artifacts_dir, f.name)
            source.append(f'args[{repr(f.name)}] = {repr(path)}')
            artifact = meta
            if not isinstance(artifact, dflow.OutputArtifact):
                artifact = dflow.OutputArtifact(path=path, archive=default_archive)  # type: ignore
            dflow_output_artifacts[f.name] = artifact

    source.extend([
        '',
        '# dump args to file',
        'with open(args_file, "w") as fp:',
        '    json.dump(args, fp, indent=2)',
        '',
        '# unpack tarball in pkg_dir',
        'for file in os.listdir(pkg_dir):',
        '    if file.endswith(".tar.bz2"):',
        '        with tarfile.open(os.path.join(pkg_dir, file), "r:bz2") as tar_fp:',
        '            tar_fp.extractall(pkg_dir)',
        '',
        '# insert pkg_dir to PYTHONPATH',
        'os.environ["PYTHONPATH"] = pkg_dir + ":" + os.environ.get("PYTHONPATH", "")',
        f'exit_code = os.system(f"{python_cmd} {{script_path}} {{args_file}}")',
        'assert exit_code == 0, f"python script failed with exit code {exit_code}"',
        eof,
    ])

    fn_str = [
        'import cloudpickle as cp',
        'import base64, json, bz2, os, sys',
        '',
        '# deserialize function',
        f'__fn = {pickle_converts(py_fn)}',
        '',
        '# deserialize args type',
        f'__ArgsType = {pickle_converts(args_type)}',
        '',
        '# run the function',
        'with open(sys.argv[1], "r") as fp:',
        '    __args = __ArgsType(**json.load(fp))',
        '__ret = __fn(__args)',
        '',
        '# handle the return value',
    ]

    for f, v in iter_python_step_return(return_type):
        path = os.path.join(output_parameters_dir, f.name)
        if _is_str_type(f.type.__origin__):
            fn_str.extend([
                f'with open({repr(path)}, "w") as fp:',
                f'    fp.write(str(__ret.{f.name}))'
            ])
        else:
            fn_str.extend([
                f'with open({repr(path)}, "w") as fp:',
                f'    json.dump(__ret.{f.name}, fp)'
            ])
        dflow_output_parameters[f.name] = dflow.OutputParameter(value_from_path=path)

    _dflow_script_check(fn_str, base_dir)
    return _PythonTemplate(source='\n'.join(source),
                           fn_str='\n'.join(fn_str),
                           script_path=script_path,
                           pkg_dir=pkg_dir,
                           dflow_input_parameters=dflow_input_parameters,
                           dflow_input_artifacts=dflow_input_artifacts,
                           dflow_output_parameters=dflow_output_parameters,
                           dflow_output_artifacts=dflow_output_artifacts)


def _s3_copy_fn(src, *args, **kwargs):
    if os.path.isdir(src):
        shutil.copytree(src, *args, dirs_exist_ok=True, **kwargs)
    else:
        shutil.copy2(src, *args, **kwargs)

class DFlowBuilder:
    """
    A type friendly wrapper to build a DFlow workflow.
    """

    def __init__(self, name:str, s3_prefix: str,
                 default_executor: Optional[DispatcherExecutor] = None,
                 default_setup_script: str = '',
                 default_archive: Optional[str] = 'default',
                 debug=False,
                 container_base_dir: str = '/tmp/dflow-builder',
                 allow_abs_s3_url=False,
                 s3_debug_fn = _s3_copy_fn):
        """
        :param name: The name of the workflow.
        :param s3_prefix: The base prefix of the S3 bucket to store data generated by the workflow.
        :param default_archive: The default archive method to use for Input/Output artifacts.
        :param default_executor: The default executor to run the workflow.
        :param default_setup_script: The default bash script to run at the beginning of each step.
        :param debug: If True, the workflow will be run in debug mode.
        :param local_mode: If True, the workflow will be run in local mode.
        :param container_base_dir: The base directory to mapping resources in remote container.
        :param allow_abs_s3_url: If True, allow absolute s3 url in input artifacts
        :param s3_debug_fn: The function to upload file to S3 under debug mode.
        """
        if debug:
            dflow.config['mode'] = 'debug'
            # dflow.config['debug_copy_method'] =  'copy'
            s3_prefix = s3_prefix.lstrip('/')

        assert container_base_dir.startswith('/tmp'), 'dflow: container_base_dir must start with /tmp'

        self.name: Final[str] = name
        self.workflow: Final[dflow.Workflow] = dflow.Workflow(name=name)
        self.s3_base_prefix: Final[str] = s3_prefix
        self.container_base_dir: Final[str] = container_base_dir

        self._default_archive = default_archive
        self._default_executor = default_executor
        self._default_setup_script = default_setup_script
        self._python_fns: Dict[Callable, str] = {}
        self._python_pkgs: Dict[str, str] = {}
        self._bash_scripts: Dict[Callable, str] = {}
        self._s3_cache: Dict[str, str] = {}
        self._s3_debug_fn = s3_debug_fn
        self._allow_abs_s3_url = allow_abs_s3_url
        self._debug = debug

    def s3_prefix(self, key: str):
        """
        get the full s3 prefix of a key.
        """
        base_prefix = self.s3_base_prefix
        # context prefix is inject by bohrium platform
        # to isolate namespace of different job
        context_prefix = dflow.s3_config.get('prefix')
        if context_prefix and not self._debug:
            base_prefix = os.path.join(context_prefix, base_prefix).strip('/ ')

        return os.path.join(base_prefix, key).strip('/ ')

    def s3_download(self, key: str, path: str = '.',
                    recursive: bool = True,
                    skip_exists: bool = False,
                    keep_dir: bool = False,
                    ):
        """
        Download file from S3 to local.
        """
        if self._debug:
            return
        dflow.download_s3(self.s3_prefix(key), path,  # type: ignore
                          recursive=recursive, skip_exists=skip_exists, keep_dir=keep_dir,
                          debug_func=self._s3_debug_fn)

    def s3_upload(self, path: Union[os.PathLike, str], key: str, cache: bool = False) -> str:
        """
        upload local file to S3.

        :param path: The local file path.
        :param keys: The keys of the S3 object.
        """
        if not isinstance(path, str):
            path = str(path)

        prefix = self.s3_prefix(key)
        if cache and prefix in self._s3_cache:
            return self._s3_cache[prefix]
        # FIXME: the type of dflow.upload_s3 is not correct
        self._s3_cache[prefix] = dflow.upload_s3(path, prefix, debug_func=self._s3_debug_fn)  # type: ignore
        return self._s3_cache[prefix]

    def s3_dump(self, data: Union[bytes, str], key: str) -> str:
        """
        Dump data to s3.

        :param data: The bytes to upload.
        :param keys: The keys of the S3 object.
        """
        mode = 'wb' if isinstance(data, bytes) else 'w'
        with tempfile.NamedTemporaryFile(mode) as fp:
            fp.write(data)
            fp.flush()
            return self.s3_upload(Path(fp.name), key)

    def add_step(self, step: Step):
        """
        Add a step to the workflow.
        """
        self.workflow.add(step.df_step)

    def add_steps(self, steps: Steps):
        """
        Add a list of steps to the workflow. The steps will be executed in parallel.
        """
        df_steps = _to_dflow_steps(steps)
        self.workflow.add(df_steps)

    def run(self, raise_on_failed=True):
        """
        Run the workflow.
        """
        self.workflow.submit()
        try:
            self.workflow.wait()
            if self.workflow.query_status() != 'Succeeded' and raise_on_failed:
                raise RuntimeError(f'workflow {self.name} failed')
        finally:
            if self._debug:
                resolve_ln(self.s3_base_prefix, mv=True)

    def make_bash_step(self, fn: Callable[[T_ARGS], types.ListStr], /,
                       uid: Optional[str] = None,
                       setup_script: str = '',
                       with_param: Any = None,
                       executor: Optional[DispatcherExecutor] = None,
                       ) -> Callable[[T_ARGS], Step[T_ARGS, None]]:
        """
        Make a bash step from python function.

        :param fn: The python function to generate the bash script.
        :param with_param: The parameter to pass to the step.
        :param setup_script: The bash script to run at the beginning of the step.
        :return: A function to run the step.
        """
        if uid is None:
            uid = str(uuid4())
        if not setup_script:
            setup_script = self._default_setup_script
        def wrapped_fn(args: T_ARGS):
            template = self._create_bash_template(fn, uid=uid, setup_script=setup_script)
            return self._build_step('bash-step-' + uid, args, template,
                                    with_param=with_param,
                                    executor=executor)
        return wrapped_fn

    def make_python_step(self, fn: Callable[[T_ARGS], T_RESULT], /,
                         uid: Optional[str] = None,
                         setup_script: str = '',
                         with_param: Any = None,
                         pkgs: Optional[Iterable[str]] = None,
                         executor: Optional[DispatcherExecutor] = None,
                         ) -> Callable[[T_ARGS], Step[T_ARGS, T_RESULT]]:
        """
        Make a python step.

        :param fn: The python function to run in the step.
        :param with_param: The parameter to pass to the step.
        :param setup_script: The bash script to run at the beginning of the step.
        :param pkgs: The python packages to install in the step.
        :return: A function to run the step.

        Due to the design flaw of the Argo Workflow that the s3 key cannot be set as step arguments,
        for each step a dedicated template have to be created.
        Ref: https://github.com/argoproj/argo-workflows/discussions/12606#discussioncomment-8358302
        """
        if uid is None:
            uid = str(uuid4())
        if pkgs is None:
            pkgs = ['dflow_galaxy', 'dflow', 'ai2_kit', 'jsonpickle']
        if not setup_script:
            setup_script = self._default_setup_script

        def wrapped_fn(args: T_ARGS):
            template = self._create_python_template(fn, uid=uid, pkgs=pkgs, setup_script=setup_script)
            return self._build_step('py-step-' + uid, args, template,
                                    with_param=with_param,
                                    executor=executor)
        return wrapped_fn

    def _create_bash_template(self, fn: Callable, uid: str,
                              setup_script: str = '',
                              bash_cmd: str = 'bash',):
        _template = bash_build_template(fn,
                                        base_dir=self.container_base_dir,
                                        setup_script=setup_script,
                                        default_archive=self._default_archive)
        dflow_template = ScriptOPTemplate(
            name='bash-template-' + uid,
            command=bash_cmd,
            script=_template.source,
        )
        dflow_template.inputs.parameters = _template.dflow_input_parameters
        dflow_template.inputs.artifacts = _template.dflow_input_artifacts
        dflow_template.outputs.artifacts = _template.dflow_output_artifacts
        return dflow_template

    def _create_python_template(self, fn: Callable, uid: str,
                                setup_script: str = '',
                                python_cmd: str = 'python3',
                                bash_cmd: str = 'bash',
                                pkgs: Optional[Iterable[str]] = None,
                                ):
        if pkgs is None:
            pkgs = []
        _template = python_build_template(fn, base_dir=self.container_base_dir,
                                          python_cmd=python_cmd,
                                          setup_script=setup_script,
                                          default_archive=self._default_archive)
        fn_hash = hashlib.sha256(_template.fn_str.encode()).hexdigest()
        dflow_template = ScriptOPTemplate(
            name='py-template-' + uid,
            command=bash_cmd,
            script=_template.source,
        )
        dflow_template.inputs.parameters = _template.dflow_input_parameters
        dflow_template.inputs.artifacts = _template.dflow_input_artifacts
        dflow_template.outputs.parameters = _template.dflow_output_parameters
        dflow_template.outputs.artifacts = _template.dflow_output_artifacts
        # upload python script to s3
        key = self._add_python_fn(fn, _template.fn_str, fn_hash)
        dflow_template.inputs.artifacts['__fn__'] = dflow.InputArtifact(
            source=dflow.S3Artifact(key=key),
            path=_template.script_path,
        )
        # download python packages
        for pkg in pkgs:
            key = self._add_python_pkg(pkg)
            dflow_template.inputs.artifacts[pkg] = dflow.InputArtifact(
                source=dflow.S3Artifact(key=key),
                path=os.path.join(_template.pkg_dir, f'{pkg}.tar.bz2'),
            )
        return dflow_template

    def _add_python_fn(self, fn, fn_str: str, fn_hash: str):
        if fn not in self._python_fns:
            fn_prefix = self.s3_dump(fn_str, f'build-in/python/fn/{fn_hash}')
            logger.info(f'upload {fn} to {fn_prefix}')
            self._python_fns[fn] = fn_prefix
        return self._python_fns[fn]

    def _add_python_pkg(self, pkg: str):
        """
        Add a python package to the workflow.
        """
        # Find the path of target package,
        # then create a temporary tarball and upload it to s3

        if pkg not in self._python_pkgs:
            pkg_path = os.path.dirname(__import__(pkg).__file__)
            with tempfile.NamedTemporaryFile(suffix='.tar.bz2') as fp:
                with tarfile.open(fp.name, 'w:bz2') as tar_fp:
                    tar_fp.add(pkg_path, arcname=os.path.basename(pkg_path), filter=_filter_pyc_files)
                fp.flush()
                key = self.s3_upload(Path(fp.name), f'build-in/python/pkg/{pkg}.tar.bz2')
                logger.info(f'upload python pkg {pkg} to {key}')
                self._python_pkgs[pkg] = key
        return self._python_pkgs[pkg]

    def _ensure_artifact(self, url_or_obj: DFLOW_ARTIFACT) -> DFLOW_ARTIFACT:
        if not isinstance(url_or_obj, str):
            return url_or_obj
        parsed = urlparse(url_or_obj)
        if parsed.scheme == 's3':
            key = parsed.path.lstrip('/')
            assert parsed.netloc in ('', '.'), f'unsupported s3 url {url_or_obj}'
            if '.' == parsed.netloc:
                key = self.s3_prefix(key)
            elif not self._allow_abs_s3_url:
                raise ValueError(f'absolute s3 url {url_or_obj} is not allowed, use relative path instead, or set allow_abs_s3_url=True')
            if self._debug:
                key = os.path.abspath(key)
            return dflow.S3Artifact(key=key)
        raise ValueError(f'unsupported url {url_or_obj}')

    def _build_step(self, name: str, args: T_ARGS, template,
                    with_param: Any=None,
                    executor: Optional[DispatcherExecutor] = None):
        if executor is None:
            executor = self._default_executor

        parameters = {}
        artifacts = {}
        for f in iter_python_step_args(args):
            meta = f.type.__metadata__[0]
            if meta == types.Symbol.INPUT_PARAMETER:
                parameters[f.name] = f.value
            elif meta == types.Symbol.INPUT_ARTIFACT or isinstance(meta, dflow.InputArtifact):
                artifacts[f.name] = self._ensure_artifact(f.value)  # type: ignore
            elif meta == types.Symbol.OUTPUT_ARTIFACT or isinstance(meta, dflow.OutputArtifact):
                template.outputs.artifacts[f.name].save = [self._ensure_artifact(f.value)]  # type: ignore
            else:
                raise ValueError(f'unsupported type {f.type}')

        step = dflow.Step(
            name=name,
            template=template,
            with_param=with_param,
            parameters=parameters,
            artifacts=artifacts,
            executor=executor,
        )
        return Step(step)


def _filter_pyc_files(tarinfo):
    if tarinfo.name.endswith('.pyc') or tarinfo.name.endswith('__pycache__'):
        return None
    return tarinfo


def _to_dflow_steps(steps: Steps):
    if isinstance(steps, Step):
        return steps.df_step
    return [_to_dflow_steps(step) for step in steps]


def _dflow_script_check(source: Iterable[str], base_dir):
    """
    ensure the script is safe to run in dflow
    """
    for line in source:
        assert '/tmp' not in line.replace(base_dir, ''), 'dflow: script should not contain /tmp literal'
        assert '"""' not in line, 'dflow: script should not contain """'
        assert '$(pwd)' not in line, 'dflow: script should not contain $(pwd)'

def _is_str_type(cls):
    # FIXME: this may have problem in some corner case
    try:
        return issubclass(cls, str)
    except TypeError:
        return False


from typing import Optional
from copy import deepcopy

from ai2_kit.core.util import load_yaml_files, merge_dict
from ai2_kit.core.cmd import CmdGroup

from dflow_galaxy.core.dflow_builder import DFlowBuilder
from dflow_galaxy.core.util import not_none
from dflow_galaxy.core.log import get_logger

from .config import TeslaConfig, WorkflowConfig
from .domain import deepmd, lammps, model_devi, cp2k
from .domain.lib import StepSwitch, LabelApp, ExploreApp

logger = get_logger(__name__)

class RuntimeContext:
    train_url: Optional[str] = None

    explore_url: Optional[str] = None
    explore_app: ExploreApp

    screen_url: Optional[str] = None

    label_url: Optional[str] = None
    label_app: Optional[LabelApp] = None


def run_tesla(*config_files: str, s3_prefix: str, debug: bool = False, skip: bool = False, max_iters: int = 1):
    builder = build_tesla_workflow(*config_files, s3_prefix=s3_prefix, debug=debug, skip=skip, max_iters=max_iters)
    builder.run()


def build_tesla_workflow(*config_files: str, s3_prefix: str, debug: bool = False, skip: bool = False, max_iters: int = 1):
    config_raw = load_yaml_files(*config_files)
    config = TeslaConfig(**config_raw)
    config.init()

    builder = DFlowBuilder(name='tesla', s3_prefix=s3_prefix, debug=debug,
                           default_archive=None)
    step_switch = StepSwitch(skip)
    runtime_ctx = RuntimeContext()

    # ensure the existence of the placeholder files or else argo will raise 404 error
    # this can not be tested in debug mode
    builder.s3_dump('', 'init-dataset/.placeholder')
    builder.s3_dump('', 'iter-dataset/.placeholder')

    raw_workflow_cfg = config.workflow
    for iter_num in range(max_iters):
        workflow_cfg = WorkflowConfig(**raw_workflow_cfg)

        type_map = workflow_cfg.general.type_map
        mass_map = workflow_cfg.general.mass_map

        iter_str = f'{iter_num:03d}'

        # Labeling
        cp2k_cfg = workflow_cfg.label.cp2k

        if cp2k_cfg and (iter_num > 0 or cp2k_cfg.init_systems):
            step_name = f'label-cp2k-iter-{iter_str}'
            runtime_ctx.label_app = 'cp2k'
            runtime_ctx.label_url = f's3://./label-cp2k/iter/{iter_str}'
            cp2k_executor = not_none(config.executors[not_none(config.orchestration.cp2k)])
            if not step_switch.shall_skip(step_name):
                if iter_num == 0:
                    assert cp2k_cfg.init_systems, 'init_systems should not be empty for first iteration'
                    assert runtime_ctx.screen_url is None, f'explore_url should be None for iter 0, actual: {runtime_ctx.screen_url}'

                for sys_key in cp2k_cfg.init_systems:
                    sys = not_none(config.datasets[sys_key])
                    builder.s3_upload(sys.url, f'init-systems/{sys_key}', cache=True)
                cp2k.provision_cp2k(builder, step_name,
                                    config=cp2k_cfg,
                                    executor=cp2k_executor,
                                    cp2k_app=not_none(cp2k_executor.apps.cp2k),
                                    python_app=not_none(cp2k_executor.apps.python),

                                    system_url=runtime_ctx.screen_url or 's3://./init-systems',
                                    work_dir_url=runtime_ctx.label_url,

                                    init=(iter_num == 0),
                                    systems=config.datasets,)

        # Training
        deepmd_cfg = workflow_cfg.train.deepmd
        if deepmd_cfg:
            step_name = f'train-deepmd-iter-{iter_str}'
            runtime_ctx.train_url = f's3://./train-deepmd/iter/{iter_str}'
            deepmd_executor = not_none(config.executors[not_none(config.orchestration.deepmd)])

            if not step_switch.shall_skip(step_name):
                for ds_key in deepmd_cfg.init_dataset:
                    ds = not_none(config.datasets[ds_key])
                    builder.s3_upload(ds.url, f'init-dataset/{ds_key}', cache=True)  # set cache to avoid re-upload
                deepmd.provision_deepmd(builder, step_name,
                                        config=deepmd_cfg,
                                        executor=deepmd_executor,
                                        deepmd_app=not_none(deepmd_executor.apps.deepmd),
                                        python_app=not_none(deepmd_executor.apps.python),

                                        label_app=runtime_ctx.label_app,
                                        label_dir_url=runtime_ctx.label_url,

                                        init_dataset_url='s3://./init-dataset',
                                        iter_dataset_url='s3://./iter-dataset',
                                        work_dir_url=runtime_ctx.train_url,
                                        iter_str=iter_str,
                                        type_map=type_map)

        else:
            raise ValueError('No training app specified')

        # Exploration
        lammps_cfg = workflow_cfg.explore.lammps
        if lammps_cfg:
            step_name = f'explore-lammps-iter-{iter_str}'
            runtime_ctx.explore_url = f's3://./explore-lammps/iter/{iter_str}'
            runtime_ctx.explore_app = 'lammps'

            lammps_executor = not_none(config.executors[not_none(config.orchestration.lammps)])

            if not step_switch.shall_skip(step_name):
                for sys_key in lammps_cfg.systems:
                    sys = not_none(config.datasets[sys_key])
                    builder.s3_upload(sys.url, f'explore-systems/{sys_key}', cache=True)

                lammps.provision_lammps(builder, step_name,
                                        config=lammps_cfg,
                                        executor=lammps_executor,
                                        lammps_app=not_none(lammps_executor.apps.lammps),
                                        python_app=not_none(lammps_executor.apps.python),

                                        mlp_model_url=runtime_ctx.train_url,
                                        systems_url='s3://./explore-systems',
                                        work_dir_url=runtime_ctx.explore_url,
                                        type_map=type_map,
                                        mass_map=mass_map,
                                        systems=config.datasets)
        else:
            raise ValueError('No explore app specified')

        # Screening
        model_devi_cfg = workflow_cfg.screen.model_devi
        if model_devi_cfg:
            step_name = f'screen-model-devi-iter-{iter_str}'
            runtime_ctx.screen_url = f's3://./screen-model-devi/iter/{iter_str}'
            model_devi_executor = not_none(config.executors[not_none(config.orchestration.model_devi)])

            if not step_switch.shall_skip(step_name):
                model_devi.provision_model_devi(builder, step_name,
                                                config=model_devi_cfg,
                                                executor=model_devi_executor,
                                                python_app=not_none(model_devi_executor.apps.python),

                                                explore_app=runtime_ctx.explore_app,
                                                explore_data_url=runtime_ctx.explore_url,

                                                persist_data_url=runtime_ctx.screen_url,
                                                type_map=type_map)
        else:
            raise ValueError('No screen app specified')

        if workflow_cfg.update:
            if iter_num == workflow_cfg.update.until_iter:
                logger.info('Updating workflow config at iter %d', iter_num)
                # the patch is applied to the original config, not the updated one
                raw_workflow_cfg = deepcopy(config.workflow)
                raw_workflow_cfg['update'] = None  # clean the old update config
                raw_workflow_cfg = merge_dict(raw_workflow_cfg, workflow_cfg.update.patch)

    return builder




cmd_entry = CmdGroup({
    'run': run_tesla,
}, doc='TESLA workflow')



from dp.launching.typing import BaseModel, Optional
from dp.launching.typing import Int, String, Enum, Float, Boolean
from dp.launching.typing import BohriumUsername, BohriumTicket, BohriumProjectId
from dp.launching.typing import (
    DflowArgoAPIServer, DflowK8sAPIServer,
    DflowAccessToken, DflowStorageEndpoint,
    DflowStorageRepository, DflowLabels,
)

from dflow.plugins import bohrium
import dflow

from dflow_galaxy.core.log import get_logger
logger = get_logger(__name__)


class DFlowOptionsMixin:
    bh_username: Optional[BohriumUsername]
    bh_ticket: Optional[BohriumTicket]
    bh_project_id: Optional[BohriumProjectId]

    dflow_labels: Optional[DflowLabels]
    dflow_argo_api_server: Optional[DflowArgoAPIServer]
    dflow_k8s_api_server: Optional[DflowK8sAPIServer]
    dflow_access_token: Optional[DflowAccessToken]
    dflow_storage_endpoint: Optional[DflowStorageEndpoint]
    dflow_storage_repository: Optional[DflowStorageRepository]


def setup_dflow_context(opts: DFlowOptionsMixin):
    """
    setup dflow context based on:
    https://dptechnology.feishu.cn/docx/HYjmdDj36oAksixbviKcbgUinUf
    """

    dflow_config = {
        'host': opts.dflow_argo_api_server,
        "k8s_api_server": opts.dflow_k8s_api_server,
        "token": opts.dflow_access_token,
        "dflow_labels": opts.dflow_labels,
    }
    dflow.config.update(dflow_config)
    logger.info(f"dflow config: {dflow.config}")

    dflow_s3_config = {
        'endpoint': opts.dflow_storage_endpoint,
        'repo_key': opts.dflow_storage_repository,
    }
    dflow.s3_config.update(dflow_s3_config)

    bohrium_config = {
        'username': opts.bh_username,
        'ticket': opts.bh_ticket,
        'project_id': opts.bh_project_id,
    }
    bohrium.config.update(bohrium_config)
    logger.info(f"bohrium config: {bohrium.config}")

    bohrium.config["tiefblue_url"] = "https://tiefblue.dp.tech"
    bohrium.config["bohrium_url"] = "https://bohrium.dp.tech"
    # override config? I have no idea
    dflow.s3_config["repo_key"] = "oss-bohrium"
    # side effect alert!!!
    # the following must be set at the end of all config
    dflow.s3_config['storage_client'] = bohrium.TiefblueClient()
    logger.info(f's3_config: {dflow.s3_config}')


class EnsembleOptions(String, Enum):
    nvt = 'nvt'
    nvt_i = 'nvt-i'
    nvt_a = 'nvt-a'
    nvt_iso = 'nvt-iso'
    nvt_aniso = 'nvt-aniso'
    npt = 'npt'
    npt_t = 'npt-t'
    npt_tri = 'npt-tri'
    nve = 'nve'
    csvr = 'csvr'